# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Endpoint(resource.Resource):
    resource_key = 'endpoint'
    resources_key = 'endpoints'
    base_path = '/endpoints'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    commit_method = 'PATCH'

    _query_mapping = resource.QueryParameters(
        'interface',
        'region_id',
        'service_id',
    )

    # Properties
    #: Describes the interface of the endpoint according to one of the
    #: following values:
    #:
    #: - `public`: intended for consumption by end users, generally on a
    #:     publicly available network interface
    #: - `internal`: not intended for consumption by end users, generally on an
    #:     unmetered internal network interface
    #: - `admin`: intended only for consumption by those needing administrative
    #:     access to the service, generally on a secure network interface
    #:
    #: *Type: string*
    interface = resource.Body('interface')
    #: Setting this value to ``False`` prevents the endpoint from appearing
    #: in the service catalog. *Type: bool*
    is_enabled = resource.Body('enabled', type=bool)
    #: The links for the region resource.
    links = resource.Body('links')
    #: Represents the containing region ID of the service endpoint.
    #: *New in v3.2* *Type: string*
    region_id = resource.Body('region_id')
    #: References the service ID to which the endpoint belongs. *Type: string*
    service_id = resource.Body('service_id')
    #: Fully qualified URL of the service endpoint. *Type: string*
    url = resource.Body('url')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.clustering.v1 import policy_type
from openstack.tests.unit import base


FAKE = {
    'name': 'FAKE_POLICY_TYPE',
    'schema': {'foo': 'bar'},
    'support_status': {'1.0': [{'status': 'supported', 'since': '2016.10'}]},
}


class TestPolicyType(base.TestCase):
    def test_basic(self):
        sot = policy_type.PolicyType()
        self.assertEqual('policy_type', sot.resource_key)
        self.assertEqual('policy_types', sot.resources_key)
        self.assertEqual('/policy-types', sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_list)

    def test_instantiate(self):
        sot = policy_type.PolicyType(**FAKE)
        self.assertEqual(FAKE['name'], sot._get_id(sot))
        self.assertEqual(FAKE['name'], sot.name)
        self.assertEqual(FAKE['schema'], sot.schema)
        self.assertEqual(FAKE['support_status'], sot.support_status)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional.shared_file_system import base


class ShareSnapshotInstanceTest(base.BaseSharedFileSystemTest):
    def setUp(self):
        super(ShareSnapshotInstanceTest, self).setUp()

        self.SHARE_NAME = self.getUniqueString()
        my_share = self.create_share(
            name=self.SHARE_NAME,
            size=2,
            share_type="dhss_false",
            share_protocol='NFS',
            description=None,
        )
        self.SHARE_ID = my_share.id
        self.create_share_snapshot(share_id=self.SHARE_ID)

    def test_share_snapshot_instances(self):
        sots = (
            self.operator_cloud.shared_file_system.share_snapshot_instances()
        )
        self.assertGreater(len(list(sots)), 0)
        for sot in sots:
            for attribute in ('id', 'name', 'created_at', 'updated_at'):
                self.assertTrue(hasattr(sot, attribute))
                self.assertIsInstance(getattr(sot, attribute), 'str')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v3 import volume as volume_
from openstack.compute.v2 import server as server_
from openstack.compute.v2 import volume_attachment as volume_attachment_
from openstack.tests.functional.compute import base as ft_base


class TestServerVolumeAttachment(ft_base.BaseComputeTest):
    def setUp(self):
        super().setUp()

        if not self.user_cloud.has_service('block-storage'):
            self.skipTest('block-storage service not supported by cloud')

        self.server_name = self.getUniqueString()
        self.volume_name = self.getUniqueString()

        # create the server and volume

        server = self.user_cloud.compute.create_server(
            name=self.server_name,
            flavor_id=self.flavor.id,
            image_id=self.image.id,
            networks='none',
        )
        self.user_cloud.compute.wait_for_server(
            server,
            wait=self._wait_for_timeout,
        )
        self.assertIsInstance(server, server_.Server)
        self.assertEqual(self.server_name, server.name)

        volume = self.user_cloud.block_storage.create_volume(
            name=self.volume_name,
            size=1,
        )
        self.user_cloud.block_storage.wait_for_status(
            volume,
            status='available',
            wait=self._wait_for_timeout,
        )
        self.assertIsInstance(volume, volume_.Volume)
        self.assertEqual(self.volume_name, volume.name)

        self.server = server
        self.volume = volume

    def tearDown(self):
        self.user_cloud.compute.delete_server(self.server.id)
        self.user_cloud.compute.wait_for_delete(
            self.server,
            wait=self._wait_for_timeout,
        )

        self.user_cloud.block_storage.delete_volume(self.volume.id)
        self.user_cloud.block_storage.wait_for_delete(
            self.volume,
            wait=self._wait_for_timeout,
        )

        super().tearDown()

    def test_volume_attachment(self):
        # create the volume attachment

        volume_attachment = self.user_cloud.compute.create_volume_attachment(
            self.server,
            self.volume,
        )
        self.assertIsInstance(
            volume_attachment,
            volume_attachment_.VolumeAttachment,
        )
        self.user_cloud.block_storage.wait_for_status(
            self.volume,
            status='in-use',
            wait=self._wait_for_timeout,
        )

        # list all attached volume attachments (there should only be one)

        volume_attachments = list(
            self.user_cloud.compute.volume_attachments(self.server)
        )
        self.assertEqual(1, len(volume_attachments))
        self.assertIsInstance(
            volume_attachments[0],
            volume_attachment_.VolumeAttachment,
        )

        # update the volume attachment

        volume_attachment = self.user_cloud.compute.update_volume_attachment(
            self.server,
            self.volume,
            delete_on_termination=True,
        )
        self.assertIsInstance(
            volume_attachment,
            volume_attachment_.VolumeAttachment,
        )

        # retrieve details of the (updated) volume attachment

        volume_attachment = self.user_cloud.compute.get_volume_attachment(
            self.server,
            self.volume,
        )
        self.assertIsInstance(
            volume_attachment,
            volume_attachment_.VolumeAttachment,
        )
        self.assertTrue(volume_attachment.delete_on_termination)

        # delete the volume attachment

        result = self.user_cloud.compute.delete_volume_attachment(
            self.server,
            self.volume,
            ignore_missing=False,
        )
        self.assertIsNone(result)

        self.user_cloud.block_storage.wait_for_status(
            self.volume,
            status='available',
            wait=self._wait_for_timeout,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.baremetal.v1 import _proxy
from openstack.baremetal.v1 import allocation
from openstack.baremetal.v1 import chassis
from openstack.baremetal.v1 import driver
from openstack.baremetal.v1 import node
from openstack.baremetal.v1 import port
from openstack.baremetal.v1 import port_group
from openstack.baremetal.v1 import volume_connector
from openstack.baremetal.v1 import volume_target
from openstack import exceptions
from openstack.tests.unit import base
from openstack.tests.unit import test_proxy_base


_MOCK_METHOD = 'openstack.baremetal.v1._proxy.Proxy._get_with_fields'


class TestBaremetalProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super(TestBaremetalProxy, self).setUp()
        self.proxy = _proxy.Proxy(self.session)


class TestDrivers(TestBaremetalProxy):
    def test_drivers(self):
        self.verify_list(self.proxy.drivers, driver.Driver)

    def test_get_driver(self):
        self.verify_get(self.proxy.get_driver, driver.Driver)


class TestChassis(TestBaremetalProxy):
    @mock.patch.object(chassis.Chassis, 'list')
    def test_chassis_detailed(self, mock_list):
        result = self.proxy.chassis(details=True, query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, details=True, query=1)

    @mock.patch.object(chassis.Chassis, 'list')
    def test_chassis_not_detailed(self, mock_list):
        result = self.proxy.chassis(query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, details=False, query=1)

    def test_create_chassis(self):
        self.verify_create(self.proxy.create_chassis, chassis.Chassis)

    def test_find_chassis(self):
        self.verify_find(self.proxy.find_chassis, chassis.Chassis)

    def test_get_chassis(self):
        self.verify_get(
            self.proxy.get_chassis,
            chassis.Chassis,
            mock_method=_MOCK_METHOD,
            expected_kwargs={'fields': None},
        )

    def test_update_chassis(self):
        self.verify_update(self.proxy.update_chassis, chassis.Chassis)

    def test_delete_chassis(self):
        self.verify_delete(self.proxy.delete_chassis, chassis.Chassis, False)

    def test_delete_chassis_ignore(self):
        self.verify_delete(self.proxy.delete_chassis, chassis.Chassis, True)


class TestNode(TestBaremetalProxy):
    @mock.patch.object(node.Node, 'list')
    def test_nodes_detailed(self, mock_list):
        result = self.proxy.nodes(details=True, query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, details=True, query=1)

    @mock.patch.object(node.Node, 'list')
    def test_nodes_not_detailed(self, mock_list):
        result = self.proxy.nodes(query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, details=False, query=1)

    @mock.patch.object(node.Node, 'list')
    def test_nodes_sharded(self, mock_list):
        kwargs = {"shard": 'meow', "query": 1}
        result = self.proxy.nodes(fields=("uuid", "instance_uuid"), **kwargs)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(
            self.proxy,
            details=False,
            fields=('uuid', 'instance_uuid'),
            shard='meow',
            query=1,
        )

    def test_create_node(self):
        self.verify_create(self.proxy.create_node, node.Node)

    def test_find_node(self):
        self.verify_find(self.proxy.find_node, node.Node)

    def test_get_node(self):
        self.verify_get(
            self.proxy.get_node,
            node.Node,
            mock_method=_MOCK_METHOD,
            expected_kwargs={'fields': None},
        )

    @mock.patch.object(node.Node, 'commit', autospec=True)
    def test_update_node(self, mock_commit):
        self.proxy.update_node('uuid', instance_id='new value')
        mock_commit.assert_called_once_with(
            mock.ANY, self.proxy, retry_on_conflict=True
        )
        self.assertEqual('new value', mock_commit.call_args[0][0].instance_id)

    @mock.patch.object(node.Node, 'commit', autospec=True)
    def test_update_node_no_retries(self, mock_commit):
        self.proxy.update_node(
            'uuid', instance_id='new value', retry_on_conflict=False
        )
        mock_commit.assert_called_once_with(
            mock.ANY, self.proxy, retry_on_conflict=False
        )
        self.assertEqual('new value', mock_commit.call_args[0][0].instance_id)

    def test_delete_node(self):
        self.verify_delete(self.proxy.delete_node, node.Node, False)

    def test_delete_node_ignore(self):
        self.verify_delete(self.proxy.delete_node, node.Node, True)


class TestPort(TestBaremetalProxy):
    @mock.patch.object(port.Port, 'list')
    def test_ports_detailed(self, mock_list):
        result = self.proxy.ports(details=True, query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, details=True, query=1)

    @mock.patch.object(port.Port, 'list')
    def test_ports_not_detailed(self, mock_list):
        result = self.proxy.ports(query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, details=False, query=1)

    def test_create_port(self):
        self.verify_create(self.proxy.create_port, port.Port)

    def test_find_port(self):
        self.verify_find(self.proxy.find_port, port.Port)

    def test_get_port(self):
        self.verify_get(
            self.proxy.get_port,
            port.Port,
            mock_method=_MOCK_METHOD,
            expected_kwargs={'fields': None},
        )

    def test_update_port(self):
        self.verify_update(self.proxy.update_port, port.Port)

    def test_delete_port(self):
        self.verify_delete(self.proxy.delete_port, port.Port, False)

    def test_delete_port_ignore(self):
        self.verify_delete(self.proxy.delete_port, port.Port, True)


class TestPortGroups(TestBaremetalProxy):
    @mock.patch.object(port_group.PortGroup, 'list')
    def test_port_groups_detailed(self, mock_list):
        result = self.proxy.port_groups(details=True, query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, details=True, query=1)

    @mock.patch.object(port_group.PortGroup, 'list')
    def test_port_groups_not_detailed(self, mock_list):
        result = self.proxy.port_groups(query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, details=False, query=1)

    def test_get_port_group(self):
        self.verify_get(
            self.proxy.get_port_group,
            port_group.PortGroup,
            mock_method=_MOCK_METHOD,
            expected_kwargs={'fields': None},
        )


class TestAllocation(TestBaremetalProxy):
    def test_create_allocation(self):
        self.verify_create(self.proxy.create_allocation, allocation.Allocation)

    def test_get_allocation(self):
        self.verify_get(
            self.proxy.get_allocation,
            allocation.Allocation,
            mock_method=_MOCK_METHOD,
            expected_kwargs={'fields': None},
        )

    def test_delete_allocation(self):
        self.verify_delete(
            self.proxy.delete_allocation, allocation.Allocation, False
        )

    def test_delete_allocation_ignore(self):
        self.verify_delete(
            self.proxy.delete_allocation, allocation.Allocation, True
        )


class TestVolumeConnector(TestBaremetalProxy):
    def test_create_volume_connector(self):
        self.verify_create(
            self.proxy.create_volume_connector,
            volume_connector.VolumeConnector,
        )

    def test_find_volume_connector(self):
        self.verify_find(
            self.proxy.find_volume_connector, volume_connector.VolumeConnector
        )

    def test_get_volume_connector(self):
        self.verify_get(
            self.proxy.get_volume_connector,
            volume_connector.VolumeConnector,
            mock_method=_MOCK_METHOD,
            expected_kwargs={'fields': None},
        )

    def test_delete_volume_connector(self):
        self.verify_delete(
            self.proxy.delete_volume_connector,
            volume_connector.VolumeConnector,
            False,
        )

    def test_delete_volume_connector_ignore(self):
        self.verify_delete(
            self.proxy.delete_volume_connector,
            volume_connector.VolumeConnector,
            True,
        )


class TestVolumeTarget(TestBaremetalProxy):
    @mock.patch.object(volume_target.VolumeTarget, 'list')
    def test_volume_target_detailed(self, mock_list):
        result = self.proxy.volume_targets(details=True, query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, detail=True, query=1)

    @mock.patch.object(volume_target.VolumeTarget, 'list')
    def test_volume_target_not_detailed(self, mock_list):
        result = self.proxy.volume_targets(query=1)
        self.assertIs(result, mock_list.return_value)
        mock_list.assert_called_once_with(self.proxy, query=1)

    def test_create_volume_target(self):
        self.verify_create(
            self.proxy.create_volume_target, volume_target.VolumeTarget
        )

    def test_find_volume_target(self):
        self.verify_find(
            self.proxy.find_volume_target, volume_target.VolumeTarget
        )

    def test_get_volume_target(self):
        self.verify_get(
            self.proxy.get_volume_target,
            volume_target.VolumeTarget,
            mock_method=_MOCK_METHOD,
            expected_kwargs={'fields': None},
        )

    def test_delete_volume_target(self):
        self.verify_delete(
            self.proxy.delete_volume_target, volume_target.VolumeTarget, False
        )

    def test_delete_volume_target_ignore(self):
        self.verify_delete(
            self.proxy.delete_volume_target, volume_target.VolumeTarget, True
        )


class TestMisc(TestBaremetalProxy):
    @mock.patch.object(node.Node, 'fetch', autospec=True)
    def test__get_with_fields_none(self, mock_fetch):
        result = self.proxy._get_with_fields(node.Node, 'value')
        self.assertIs(result, mock_fetch.return_value)
        mock_fetch.assert_called_once_with(
            mock.ANY, self.proxy, error_message=mock.ANY
        )

    @mock.patch.object(node.Node, 'fetch', autospec=True)
    def test__get_with_fields_node(self, mock_fetch):
        result = self.proxy._get_with_fields(
            # Mix of server-side and client-side fields
            node.Node,
            'value',
            fields=['maintenance', 'id', 'instance_id'],
        )
        self.assertIs(result, mock_fetch.return_value)
        mock_fetch.assert_called_once_with(
            mock.ANY,
            self.proxy,
            error_message=mock.ANY,
            # instance_id converted to server-side instance_uuid
            fields='maintenance,uuid,instance_uuid',
        )

    @mock.patch.object(port.Port, 'fetch', autospec=True)
    def test__get_with_fields_port(self, mock_fetch):
        result = self.proxy._get_with_fields(
            port.Port, 'value', fields=['address', 'id', 'node_id']
        )
        self.assertIs(result, mock_fetch.return_value)
        mock_fetch.assert_called_once_with(
            mock.ANY,
            self.proxy,
            error_message=mock.ANY,
            # node_id converted to server-side node_uuid
            fields='address,uuid,node_uuid',
        )


@mock.patch('time.sleep', lambda _sec: None)
@mock.patch.object(_proxy.Proxy, 'get_node', autospec=True)
class TestWaitForNodesProvisionState(base.TestCase):
    def setUp(self):
        super(TestWaitForNodesProvisionState, self).setUp()
        self.session = mock.Mock()
        self.proxy = _proxy.Proxy(self.session)

    def test_success(self, mock_get):
        # two attempts, one node succeeds after the 1st
        nodes = [mock.Mock(spec=node.Node, id=str(i)) for i in range(3)]
        for i, n in enumerate(nodes):
            # 1st attempt on 1st node, 2nd attempt on 2nd node
            n._check_state_reached.return_value = not (i % 2)
            mock_get.side_effect = nodes

        result = self.proxy.wait_for_nodes_provision_state(
            ['abcd', node.Node(id='1234')], 'fake state'
        )
        self.assertEqual([nodes[0], nodes[2]], result)

        for n in nodes:
            n._check_state_reached.assert_called_once_with(
                self.proxy, 'fake state', True
            )

    def test_success_no_fail(self, mock_get):
        # two attempts, one node succeeds after the 1st
        nodes = [mock.Mock(spec=node.Node, id=str(i)) for i in range(3)]
        for i, n in enumerate(nodes):
            # 1st attempt on 1st node, 2nd attempt on 2nd node
            n._check_state_reached.return_value = not (i % 2)
            mock_get.side_effect = nodes

        result = self.proxy.wait_for_nodes_provision_state(
            ['abcd', node.Node(id='1234')], 'fake state', fail=False
        )
        self.assertEqual([nodes[0], nodes[2]], result.success)
        self.assertEqual([], result.failure)
        self.assertEqual([], result.timeout)

        for n in nodes:
            n._check_state_reached.assert_called_once_with(
                self.proxy, 'fake state', True
            )

    def test_timeout(self, mock_get):
        mock_get.return_value._check_state_reached.return_value = False
        mock_get.return_value.id = '1234'

        self.assertRaises(
            exceptions.ResourceTimeout,
            self.proxy.wait_for_nodes_provision_state,
            ['abcd', node.Node(id='1234')],
            'fake state',
            timeout=0.001,
        )
        mock_get.return_value._check_state_reached.assert_called_with(
            self.proxy, 'fake state', True
        )

    def test_timeout_no_fail(self, mock_get):
        mock_get.return_value._check_state_reached.return_value = False
        mock_get.return_value.id = '1234'

        result = self.proxy.wait_for_nodes_provision_state(
            ['abcd'], 'fake state', timeout=0.001, fail=False
        )
        mock_get.return_value._check_state_reached.assert_called_with(
            self.proxy, 'fake state', True
        )

        self.assertEqual([], result.success)
        self.assertEqual([mock_get.return_value], result.timeout)
        self.assertEqual([], result.failure)

    def test_timeout_and_failures_not_fail(self, mock_get):
        def _fake_get(_self, node):
            result = mock.Mock()
            result.id = getattr(node, 'id', node)
            if result.id == '1':
                result._check_state_reached.return_value = True
            elif result.id == '2':
                result._check_state_reached.side_effect = (
                    exceptions.ResourceFailure("boom")
                )
            else:
                result._check_state_reached.return_value = False
            return result

        mock_get.side_effect = _fake_get

        result = self.proxy.wait_for_nodes_provision_state(
            ['1', '2', '3'], 'fake state', timeout=0.001, fail=False
        )

        self.assertEqual(['1'], [x.id for x in result.success])
        self.assertEqual(['3'], [x.id for x in result.timeout])
        self.assertEqual(['2'], [x.id for x in result.failure])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import functools
import operator
import time

import iso8601

from openstack.cloud import _utils
from openstack.cloud import exc
from openstack.cloud import meta
from openstack.compute.v2._proxy import Proxy
from openstack.compute.v2 import quota_set as _qs
from openstack.compute.v2 import server as _server
from openstack import exceptions
from openstack import utils


_SERVER_FIELDS = (
    'accessIPv4',
    'accessIPv6',
    'addresses',
    'adminPass',
    'created',
    'description',
    'key_name',
    'metadata',
    'networks',
    'personality',
    'private_v4',
    'public_v4',
    'public_v6',
    'server_groups',
    'status',
    'updated',
    'user_id',
    'tags',
)


def _to_bool(value):
    if isinstance(value, str):
        if not value:
            return False
        prospective = value.lower().capitalize()
        return prospective == 'True'
    return bool(value)


def _pop_int(resource, key):
    return int(resource.pop(key, 0) or 0)


def _pop_or_get(resource, key, default, strict):
    if strict:
        return resource.pop(key, default)
    else:
        return resource.get(key, default)


class ComputeCloudMixin:
    compute: Proxy

    @property
    def _compute_region(self):
        # This is only used in exception messages. Can we get rid of it?
        return self.config.get_region_name('compute')

    def get_flavor_name(self, flavor_id):
        """Get the name of a flavor.

        :param flavor_id: ID of the flavor.
        :returns: The name of the flavor if a match if found, else None.
        """
        flavor = self.get_flavor(flavor_id, get_extra=False)
        if flavor:
            return flavor['name']
        return None

    def get_flavor_by_ram(self, ram, include=None, get_extra=True):
        """Get a flavor based on amount of RAM available.

        Finds the flavor with the least amount of RAM that is at least
        as much as the specified amount. If `include` is given, further
        filter based on matching flavor name.

        :param int ram: Minimum amount of RAM.
        :param string include: If given, will return a flavor whose name
            contains this string as a substring.
        :param get_extra:

        :returns: A compute ``Flavor`` object.
        :raises: :class:`~openstack.exceptions.SDKException` if no
            matching flavour could be found.
        """
        flavors = self.list_flavors(get_extra=get_extra)
        for flavor in sorted(flavors, key=operator.itemgetter('ram')):
            if flavor['ram'] >= ram and (
                not include or include in flavor['name']
            ):
                return flavor
        raise exceptions.SDKException(
            "Could not find a flavor with {ram} and '{include}'".format(
                ram=ram, include=include
            )
        )

    def _nova_extensions(self):
        extensions = set([e.alias for e in self.compute.extensions()])
        return extensions

    def _has_nova_extension(self, extension_name):
        return extension_name in self._nova_extensions()

    def search_keypairs(self, name_or_id=None, filters=None):
        """Search keypairs.

        :param name_or_id:
        :param filters:
        :returns: A list of compute ``Keypair`` objects matching the search
            criteria.
        """
        keypairs = self.list_keypairs(
            filters=filters if isinstance(filters, dict) else None
        )
        return _utils._filter_list(keypairs, name_or_id, filters)

    def search_flavors(self, name_or_id=None, filters=None, get_extra=True):
        """Search flavors.

        :param name_or_id:
        :param flavors:
        :param get_extra:
        :returns: A list of compute ``Flavor`` objects matching the search
            criteria.
        """
        flavors = self.list_flavors(get_extra=get_extra)
        return _utils._filter_list(flavors, name_or_id, filters)

    def search_servers(
        self,
        name_or_id=None,
        filters=None,
        detailed=False,
        all_projects=False,
        bare=False,
    ):
        """Search servers.

        :param name_or_id:
        :param filters:
        :param detailed:
        :param all_projects:
        :param bare:
        :returns: A list of compute ``Server`` objects matching the search
            criteria.
        """
        servers = self.list_servers(
            detailed=detailed, all_projects=all_projects, bare=bare
        )
        return _utils._filter_list(servers, name_or_id, filters)

    def search_server_groups(self, name_or_id=None, filters=None):
        """Search server groups.

        :param name_or_id: Name or unique ID of the server group(s).
        :param filters: A dict containing additional filters to use.

        :returns: A list of compute ``ServerGroup`` objects matching the search
            criteria.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        server_groups = self.list_server_groups()
        return _utils._filter_list(server_groups, name_or_id, filters)

    def list_keypairs(self, filters=None):
        """List all available keypairs.

        :param filters:
        :returns: A list of compute ``Keypair`` objects.
        """
        if not filters:
            filters = {}
        return list(self.compute.keypairs(**filters))

    def list_availability_zone_names(self, unavailable=False):
        """List names of availability zones.

        :param bool unavailable: Whether or not to include unavailable zones
            in the output. Defaults to False.
        :returns: A list of availability zone names, or an empty list if the
            list could not be fetched.
        """
        try:
            zones = self.compute.availability_zones()
            ret = []
            for zone in zones:
                if zone.state['available'] or unavailable:
                    ret.append(zone.name)
            return ret
        except exceptions.SDKException:
            self.log.debug(
                "Availability zone list could not be fetched", exc_info=True
            )
            return []

    def list_flavors(self, get_extra=False):
        """List all available flavors.

        :param get_extra: Whether or not to fetch extra specs for each flavor.
            Defaults to True. Default behavior value can be overridden in
            clouds.yaml by setting openstack.cloud.get_extra_specs to False.
        :returns: A list of compute ``Flavor`` objects.
        """
        return list(
            self.compute.flavors(details=True, get_extra_specs=get_extra)
        )

    def list_server_security_groups(self, server):
        """List all security groups associated with the given server.

        :returns: A list of security group dictionary objects.
        """

        # Don't even try if we're a cloud that doesn't have them
        if not self._has_secgroups():
            return []

        server = self.compute.get_server(server)

        server.fetch_security_groups(self.compute)

        return server.security_groups

    def _get_server_security_groups(self, server, security_groups):
        if not self._has_secgroups():
            raise exc.OpenStackCloudUnavailableFeature(
                "Unavailable feature: security groups"
            )

        if not isinstance(server, dict):
            server = self.get_server(server, bare=True)

            if server is None:
                self.log.debug('Server %s not found', server)
                return None, None

        if not isinstance(security_groups, (list, tuple)):
            security_groups = [security_groups]

        sec_group_objs = []

        for sg in security_groups:
            if not isinstance(sg, dict):
                sg = self.get_security_group(sg)

                if sg is None:
                    self.log.debug(
                        'Security group %s not found for adding', sg
                    )

                    return None, None

            sec_group_objs.append(sg)

        return server, sec_group_objs

    def add_server_security_groups(self, server, security_groups):
        """Add security groups to a server.

        Add existing security groups to an existing server. If the security
        groups are already present on the server this will continue unaffected.

        :returns: False if server or security groups are undefined, True
            otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        server, security_groups = self._get_server_security_groups(
            server, security_groups
        )

        if not (server and security_groups):
            return False

        for sg in security_groups:
            self.compute.add_security_group_to_server(server, sg)

        return True

    def remove_server_security_groups(self, server, security_groups):
        """Remove security groups from a server

        Remove existing security groups from an existing server. If the
        security groups are not present on the server this will continue
        unaffected.

        :returns: False if server or security groups are undefined, True
            otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        server, security_groups = self._get_server_security_groups(
            server, security_groups
        )

        if not (server and security_groups):
            return False

        ret = True

        for sg in security_groups:
            try:
                self.compute.remove_security_group_from_server(server, sg)

            except exceptions.ResourceNotFound:
                # NOTE(jamielennox): Is this ok? If we remove something that
                # isn't present should we just conclude job done or is that an
                # error? Nova returns ok if you try to add a group twice.
                self.log.debug(
                    "The security group %s was not present on server %s so "
                    "no action was performed",
                    sg.name,
                    server.name,
                )
                ret = False

        return ret

    def list_servers(
        self,
        detailed=False,
        all_projects=False,
        bare=False,
        filters=None,
    ):
        """List all available servers.

        :param detailed: Whether or not to add detailed additional information.
            Defaults to False.
        :param all_projects: Whether to list servers from all projects or just
            the current auth scoped project.
        :param bare: Whether to skip adding any additional information to the
            server record. Defaults to False, meaning the addresses dict will
            be populated as needed from neutron. Setting to True implies
            detailed = False.
        :param filters: Additional query parameters passed to the API server.
        :returns: A list of compute ``Server`` objects.
        """
        if not filters:
            filters = {}

        return [
            self._expand_server(server, detailed, bare)
            for server in self.compute.servers(
                all_projects=all_projects,
                **filters,
            )
        ]

    def list_server_groups(self):
        """List all available server groups.

        :returns: A list of compute ``ServerGroup`` objects.
        """
        return list(self.compute.server_groups())

    def get_compute_limits(self, name_or_id=None):
        """Get absolute compute limits for a project

        :param name_or_id: (optional) project name or ID to get limits for
            if different from the current project

        :returns: A compute
            :class:`~openstack.compute.v2.limits.Limits.AbsoluteLimits` object.
        :raises: :class:`~openstack.exceptions.SDKException` if it's not a
            valid project
        """
        params = {}
        project_id = None
        if name_or_id:
            proj = self.get_project(name_or_id)
            if not proj:
                raise exceptions.SDKException("project does not exist")
            project_id = proj.id
            params['tenant_id'] = project_id
        return self.compute.get_limits(**params).absolute

    def get_keypair(self, name_or_id, filters=None):
        """Get a keypair by name or ID.

        :param name_or_id: Name or ID of the keypair.
        :param filters: A dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {
                'last_name': 'Smith',
                'other': {
                'gender': 'Female'
                }
                }

            OR
            A string containing a jmespath expression for further filtering.
            Example:: "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :returns: A compute ``Keypair`` object if found, else None.
        """
        return _utils._get_entity(self, 'keypair', name_or_id, filters)

    def get_flavor(self, name_or_id, filters=None, get_extra=True):
        """Get a flavor by name or ID.

        :param name_or_id: Name or ID of the flavor.
        :param filters: A dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {
                    'last_name': 'Smith',
                    'other': {
                        'gender': 'Female'
                    }
                }

            OR
            A string containing a jmespath expression for further filtering.
            Example:: "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :param get_extra: Whether or not the list_flavors call should get the
            extra flavor specs.
        :returns: A compute ``Flavor`` object if found, else None.
        """
        if not filters:
            filters = {}
        flavor = self.compute.find_flavor(
            name_or_id,
            get_extra_specs=get_extra,
            ignore_missing=True,
            **filters,
        )
        return flavor

    def get_flavor_by_id(self, id, get_extra=False):
        """Get a flavor by ID

        :param id: ID of the flavor.
        :param get_extra: Whether or not the list_flavors call should get the
            extra flavor specs.
        :returns: A compute ``Flavor`` object if found, else None.
        """
        return self.compute.get_flavor(id, get_extra_specs=get_extra)

    def get_server_console(self, server, length=None):
        """Get the console log for a server.

        :param server: The server to fetch the console log for. Can be either
            a server dict or the Name or ID of the server.
        :param int length: The number of lines you would like to retrieve from
            the end of the log. (optional, defaults to all)

        :returns: A string containing the text of the console log or an
            empty string if the cloud does not support console logs.
        :raises: :class:`~openstack.exceptions.SDKException` if an invalid
            server argument is given or if something else unforseen happens
        """

        if not isinstance(server, dict):
            server = self.get_server(server, bare=True)

        if not server:
            raise exceptions.SDKException(
                "Console log requested for invalid server"
            )

        try:
            return self._get_server_console_output(server['id'], length)
        except exceptions.BadRequestException:
            return ""

    def _get_server_console_output(self, server_id, length=None):
        output = self.compute.get_server_console_output(
            server=server_id, length=length
        )
        if 'output' in output:
            return output['output']

    def get_server(
        self,
        name_or_id=None,
        filters=None,
        detailed=False,
        bare=False,
        all_projects=False,
    ):
        """Get a server by name or ID.

        :param name_or_id: Name or ID of the server.
        :param filters:
            A dictionary of meta data to use for further filtering. Elements
            of this dictionary may, themselves, be dictionaries. Example::

                {
                    'last_name': 'Smith',
                    'other': {
                        'gender': 'Female'
                    }
                }

            OR
            A string containing a jmespath expression for further filtering.
            Example:: "[?last_name==`Smith`] | [?other.gender]==`Female`]"
        :param detailed: Whether or not to add detailed additional information.
            Defaults to False.
        :param bare: Whether to skip adding any additional information to the
            server record. Defaults to False, meaning the addresses dict will
            be populated as needed from neutron. Setting to True implies
            detailed = False.
        :param all_projects: Whether to get server from all projects or just
            the current auth scoped project.
        :returns: A compute ``Server`` object if found, else None.
        """
        searchfunc = functools.partial(
            self.search_servers,
            detailed=detailed,
            bare=True,
            all_projects=all_projects,
        )
        server = _utils._get_entity(self, searchfunc, name_or_id, filters)
        return self._expand_server(server, detailed, bare)

    def _expand_server(self, server, detailed, bare):
        if bare or not server:
            return server
        elif detailed:
            return meta.get_hostvars_from_server(self, server)
        else:
            return meta.add_server_interfaces(self, server)

    def get_server_by_id(self, id):
        """Get a server by ID.

        :param id: ID of the server.

        :returns: A compute ``Server`` object if found, else None.
        """
        try:
            server = self.compute.get_server(id)
            return meta.add_server_interfaces(self, server)
        except exceptions.ResourceNotFound:
            return None

    def get_server_group(self, name_or_id=None, filters=None):
        """Get a server group by name or ID.

        :param name_or_id: Name or ID of the server group.
        :param filters: A dictionary of meta data to use for further filtering.
            Elements of this dictionary may, themselves, be dictionaries.
            Example::

                {
                    'policy': 'affinity',
                }

            OR
            A string containing a jmespath expression for further filtering.
            Example:: "[?last_name==`Smith`] | [?other.gender]==`Female`]"

        :returns: A compute ``ServerGroup`` object if found, else None.
        """
        return _utils._get_entity(self, 'server_group', name_or_id, filters)

    def create_keypair(self, name, public_key=None):
        """Create a new keypair.

        :param name: Name of the keypair being created.
        :param public_key: Public key for the new keypair.

        :returns: The created compute ``Keypair`` object.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        keypair = {
            'name': name,
        }
        if public_key:
            keypair['public_key'] = public_key
        return self.compute.create_keypair(**keypair)

    def delete_keypair(self, name):
        """Delete a keypair.

        :param name: Name of the keypair to delete.

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        try:
            self.compute.delete_keypair(name, ignore_missing=False)
        except exceptions.ResourceNotFound:
            self.log.debug("Keypair %s not found for deleting", name)
            return False
        return True

    def create_image_snapshot(
        self,
        name,
        server,
        wait=False,
        timeout=3600,
        **metadata,
    ):
        """Create an image by snapshotting an existing server.

        ..note::
            On most clouds this is a cold snapshot - meaning that the server in
            question will be shutdown before taking the snapshot. It is
            possible that it's a live snapshot - but there is no way to know as
            a user, so caveat emptor.

        :param name: Name of the image to be created
        :param server: Server name or ID or dict representing the server
            to be snapshotted
        :param wait: If true, waits for image to be created.
        :param timeout: Seconds to wait for image creation. None is forever.
        :param metadata: Metadata to give newly-created image entity

        :returns: The created image ``Image`` object.
        :raises: :class:`~openstack.exceptions.SDKException` if there are
            problems uploading
        """
        if not isinstance(server, dict):
            server_obj = self.get_server(server, bare=True)
            if not server_obj:
                raise exceptions.SDKException(
                    "Server {server} could not be found and therefore"
                    " could not be snapshotted.".format(server=server)
                )
            server = server_obj
        image = self.compute.create_server_image(
            server, name=name, metadata=metadata, wait=wait, timeout=timeout
        )
        return image

    def get_server_id(self, name_or_id):
        """Get the ID of a server.

        :param name_or_id:
        :returns: The name of the server if found, else None.
        """
        server = self.get_server(name_or_id, bare=True)
        if server:
            return server['id']
        return None

    def get_server_private_ip(self, server):
        """Get the private IP of a server.

        :param server:
        :returns: The private IP of the server if set, else None.
        """
        return meta.get_server_private_ip(server, self)

    def get_server_public_ip(self, server):
        """Get the public IP of a server.

        :param server:
        :returns: The public IP of the server if set, else None.
        """
        return meta.get_server_external_ipv4(self, server)

    def get_server_meta(self, server):
        """Get the metadata for a server.

        :param server:
        :returns: The metadata for the server if found, else None.
        """
        # TODO(mordred) remove once ansible has moved to Inventory interface
        server_vars = meta.get_hostvars_from_server(self, server)
        groups = meta.get_groups_from_server(self, server, server_vars)
        return dict(server_vars=server_vars, groups=groups)

    @_utils.valid_kwargs(
        'meta',
        'files',
        'userdata',
        'description',
        'reservation_id',
        'return_raw',
        'min_count',
        'max_count',
        'security_groups',
        'key_name',
        'availability_zone',
        'block_device_mapping',
        'block_device_mapping_v2',
        'nics',
        'scheduler_hints',
        'config_drive',
        'admin_pass',
        'disk_config',
    )
    def create_server(
        self,
        name,
        image=None,
        flavor=None,
        auto_ip=True,
        ips=None,
        ip_pool=None,
        root_volume=None,
        terminate_volume=False,
        wait=False,
        timeout=180,
        reuse_ips=True,
        network=None,
        boot_from_volume=False,
        volume_size='50',
        boot_volume=None,
        volumes=None,
        nat_destination=None,
        group=None,
        **kwargs,
    ):
        """Create a virtual server instance.

        :param name: Something to name the server.
        :param image: Image dict, name or ID to boot with. image is required
            unless boot_volume is given.
        :param flavor: Flavor dict, name or ID to boot onto.
        :param auto_ip: Whether to take actions to find a routable IP for
            the server. (defaults to True)
        :param ips: List of IPs to attach to the server (defaults to None)
        :param ip_pool: Name of the network or floating IP pool to get an
            address from. (defaults to None)
        :param root_volume: Name or ID of a volume to boot from
            (defaults to None - deprecated, use boot_volume)
        :param boot_volume: Name or ID of a volume to boot from
            (defaults to None)
        :param terminate_volume: If booting from a volume, whether it should
            be deleted when the server is destroyed.  (defaults to False)
        :param volumes: (optional) A list of volumes to attach to the server
        :param meta: (optional) A dict of arbitrary key/value metadata to
            store for this server. Both keys and values must be <=255
            characters.
        :param files: (optional, deprecated) A dict of files to overwrite
            on the server upon boot. Keys are file names (i.e.
            ``/etc/passwd``) and values are the file contents (either as a
            string or as a file-like object). A maximum of five entries is
            allowed, and each file must be 10k or less.
        :param reservation_id: a UUID for the set of servers being requested.
        :param min_count: (optional extension) The minimum number of servers to
            launch.
        :param max_count: (optional extension) The maximum number of servers to
            launch.
        :param security_groups: A list of security group names
        :param userdata: user data to pass to be exposed by the metadata
            server this can be a file type object as well or a string.
        :param key_name: (optional extension) name of previously created
            keypair to inject into the instance.
        :param availability_zone: Name of the availability zone for instance
            placement.
        :param block_device_mapping: (optional) A dict of block
            device mappings for this server.
        :param block_device_mapping_v2: (optional) A dict of block
            device mappings for this server.
        :param nics:  (optional extension) an ordered list of nics to be
            added to this server, with information about connected networks,
            fixed IPs, port etc.
        :param scheduler_hints: (optional extension) arbitrary key-value pairs
            specified by the client to help boot an instance
        :param config_drive: (optional extension) value for config drive
            either boolean, or volume-id
        :param disk_config: (optional extension) control how the disk is
            partitioned when the server is created. possible values are 'AUTO'
            or 'MANUAL'.
        :param admin_pass: (optional extension) add a user supplied admin
            password.
        :param wait: (optional) Wait for the address to appear as assigned
            to the server. Defaults to False.
        :param timeout: (optional) Seconds to wait, defaults to 60.
            See the ``wait`` parameter.
        :param reuse_ips: (optional) Whether to attempt to reuse pre-existing
            floating ips should a floating IP be needed (defaults to True)
        :param network: (optional) Network dict or name or ID to attach the
            server to. Mutually exclusive with the nics parameter. Can also
            be a list of network names or IDs or network dicts.
        :param boot_from_volume: Whether to boot from volume. 'boot_volume'
            implies True, but boot_from_volume=True with no boot_volume is
            valid and will create a volume from the image and use that.
        :param volume_size: When booting an image from volume, how big should
            the created volume be? Defaults to 50.
        :param nat_destination: Which network should a created floating IP
            be attached to, if it's not possible to infer from the cloud's
            configuration. (Optional, defaults to None)
        :param group: ServerGroup dict, name or id to boot the server in.
            If a group is provided in both scheduler_hints and in the group
            param, the group param will win. (Optional, defaults to None)

        :returns: The created compute ``Server`` object.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        # TODO(shade) Image is optional but flavor is not - yet flavor comes
        # after image in the argument list. Doh.
        if not flavor:
            raise TypeError(
                "create_server() missing 1 required argument: 'flavor'"
            )
        if not image and not boot_volume:
            raise TypeError(
                "create_server() requires either 'image' or 'boot_volume'"
            )

        # TODO(mordred) Add support for description starting in 2.19
        security_groups = kwargs.get('security_groups', [])
        if security_groups and not isinstance(kwargs['security_groups'], list):
            security_groups = [security_groups]
        if security_groups:
            kwargs['security_groups'] = []
            for sec_group in security_groups:
                kwargs['security_groups'].append(dict(name=sec_group))
        if 'userdata' in kwargs:
            user_data = kwargs.pop('userdata')
            if user_data:
                kwargs['user_data'] = self._encode_server_userdata(user_data)
        for desired, given in (
            ('OS-DCF:diskConfig', 'disk_config'),
            ('config_drive', 'config_drive'),
            ('key_name', 'key_name'),
            ('metadata', 'meta'),
            ('adminPass', 'admin_pass'),
        ):
            value = kwargs.pop(given, None)
            if value:
                kwargs[desired] = value

        if group:
            group_obj = self.get_server_group(group)
            if not group_obj:
                raise exceptions.SDKException(
                    "Server Group {group} was requested but was not found"
                    " on the cloud".format(group=group)
                )
            if 'scheduler_hints' not in kwargs:
                kwargs['scheduler_hints'] = {}
            kwargs['scheduler_hints']['group'] = group_obj['id']

        kwargs.setdefault('max_count', kwargs.get('max_count', 1))
        kwargs.setdefault('min_count', kwargs.get('min_count', 1))

        if 'nics' in kwargs and not isinstance(kwargs['nics'], list):
            if isinstance(kwargs['nics'], dict):
                # Be nice and help the user out
                kwargs['nics'] = [kwargs['nics']]
            else:
                raise exceptions.SDKException(
                    'nics parameter to create_server takes a list of dicts.'
                    ' Got: {nics}'.format(nics=kwargs['nics'])
                )

        if network and ('nics' not in kwargs or not kwargs['nics']):
            nics = []
            if not isinstance(network, list):
                network = [network]
            for net_name in network:
                if isinstance(net_name, dict) and 'id' in net_name:
                    network_obj = net_name
                else:
                    network_obj = self.get_network(name_or_id=net_name)
                if not network_obj:
                    raise exceptions.SDKException(
                        'Network {network} is not a valid network in'
                        ' {cloud}:{region}'.format(
                            network=network,
                            cloud=self.name,
                            region=self._compute_region,
                        )
                    )
                nics.append({'net-id': network_obj['id']})

            kwargs['nics'] = nics
        if not network and ('nics' not in kwargs or not kwargs['nics']):
            default_network = self.get_default_network()
            if default_network:
                kwargs['nics'] = [{'net-id': default_network['id']}]

        networks = []
        for nic in kwargs.pop('nics', []):
            net = {}
            if 'net-id' in nic:
                # TODO(mordred) Make sure this is in uuid format
                net['uuid'] = nic.pop('net-id')
                # If there's a net-id, ignore net-name
                nic.pop('net-name', None)
            elif 'net-name' in nic:
                net_name = nic.pop('net-name')
                nic_net = self.get_network(net_name)
                if not nic_net:
                    raise exceptions.SDKException(
                        "Requested network {net} could not be found.".format(
                            net=net_name
                        )
                    )
                net['uuid'] = nic_net['id']
            for ip_key in ('v4-fixed-ip', 'v6-fixed-ip', 'fixed_ip'):
                fixed_ip = nic.pop(ip_key, None)
                if fixed_ip and net.get('fixed_ip'):
                    raise exceptions.SDKException(
                        "Only one of v4-fixed-ip, v6-fixed-ip or fixed_ip"
                        " may be given"
                    )
                if fixed_ip:
                    net['fixed_ip'] = fixed_ip
            for key in ('port', 'port-id'):
                if key in nic:
                    net['port'] = nic.pop(key)
            # A tag supported only in server microversion 2.32-2.36 or >= 2.42
            # Bumping the version to 2.42 to support the 'tag' implementation
            if 'tag' in nic:
                utils.require_microversion(self.compute, '2.42')
                net['tag'] = nic.pop('tag')
            if nic:
                raise exceptions.SDKException(
                    "Additional unsupported keys given for server network"
                    " creation: {keys}".format(keys=nic.keys())
                )
            networks.append(net)
        if networks:
            kwargs['networks'] = networks
        else:
            # If user has not passed networks - let Nova try the best;
            # note earlier microversions expect this to be blank.
            if utils.supports_microversion(self.compute, '2.37'):
                kwargs['networks'] = 'auto'

        if image:
            if isinstance(image, dict):
                kwargs['imageRef'] = image['id']
            else:
                kwargs['imageRef'] = self.get_image(image).id
        if isinstance(flavor, dict):
            kwargs['flavorRef'] = flavor['id']
        else:
            kwargs['flavorRef'] = self.get_flavor(flavor, get_extra=False).id

        if volumes is None:
            volumes = []

        # nova cli calls this boot_volume. Let's be the same
        if root_volume and not boot_volume:
            boot_volume = root_volume

        kwargs = self._get_boot_from_volume_kwargs(
            image=image,
            boot_from_volume=boot_from_volume,
            boot_volume=boot_volume,
            volume_size=str(volume_size),
            terminate_volume=terminate_volume,
            volumes=volumes,
            kwargs=kwargs,
        )

        kwargs['name'] = name

        server = self.compute.create_server(**kwargs)
        # TODO(mordred) We're only testing this in functional tests. We need
        # to add unit tests for this too.
        admin_pass = server.admin_password or kwargs.get('admin_pass')
        if not wait:
            server = self.compute.get_server(server.id)
            if server.status == 'ERROR':
                raise exc.OpenStackCloudCreateException(
                    resource='server', resource_id=server.id
                )
            server = meta.add_server_interfaces(self, server)

        else:
            server = self.wait_for_server(
                server,
                auto_ip=auto_ip,
                ips=ips,
                ip_pool=ip_pool,
                reuse=reuse_ips,
                timeout=timeout,
                nat_destination=nat_destination,
            )

        server.admin_password = admin_pass
        return server

    def _get_boot_from_volume_kwargs(
        self,
        image,
        boot_from_volume,
        boot_volume,
        volume_size,
        terminate_volume,
        volumes,
        kwargs,
    ):
        """Return block device mappings

        :param image: Image dict, name or id to boot with.

        """
        # TODO(mordred) We're only testing this in functional tests. We need
        # to add unit tests for this too.
        if boot_volume or boot_from_volume or volumes:
            kwargs.setdefault('block_device_mapping_v2', [])
        else:
            return kwargs

        # If we have boot_from_volume but no root volume, then we're
        # booting an image from volume
        if boot_volume:
            volume = self.get_volume(boot_volume)
            if not volume:
                raise exceptions.SDKException(
                    'Volume {boot_volume} is not a valid volume'
                    ' in {cloud}:{region}'.format(
                        boot_volume=boot_volume,
                        cloud=self.name,
                        region=self._compute_region,
                    )
                )
            block_mapping = {
                'boot_index': '0',
                'delete_on_termination': terminate_volume,
                'destination_type': 'volume',
                'uuid': volume['id'],
                'source_type': 'volume',
            }
            kwargs['block_device_mapping_v2'].append(block_mapping)
            kwargs['imageRef'] = ''
        elif boot_from_volume:
            if isinstance(image, dict):
                image_obj = image
            else:
                image_obj = self.get_image(image)
            if not image_obj:
                raise exceptions.SDKException(
                    'Image {image} is not a valid image in'
                    ' {cloud}:{region}'.format(
                        image=image,
                        cloud=self.name,
                        region=self._compute_region,
                    )
                )

            block_mapping = {
                'boot_index': '0',
                'delete_on_termination': terminate_volume,
                'destination_type': 'volume',
                'uuid': image_obj['id'],
                'source_type': 'image',
                'volume_size': volume_size,
            }
            kwargs['imageRef'] = ''
            kwargs['block_device_mapping_v2'].append(block_mapping)
        if volumes and kwargs['imageRef']:
            # If we're attaching volumes on boot but booting from an image,
            # we need to specify that in the BDM.
            block_mapping = {
                u'boot_index': 0,
                u'delete_on_termination': True,
                u'destination_type': u'local',
                u'source_type': u'image',
                u'uuid': kwargs['imageRef'],
            }
            kwargs['block_device_mapping_v2'].append(block_mapping)
        for volume in volumes:
            volume_obj = self.get_volume(volume)
            if not volume_obj:
                raise exceptions.SDKException(
                    'Volume {volume} is not a valid volume'
                    ' in {cloud}:{region}'.format(
                        volume=volume,
                        cloud=self.name,
                        region=self._compute_region,
                    )
                )
            block_mapping = {
                'boot_index': '-1',
                'delete_on_termination': False,
                'destination_type': 'volume',
                'uuid': volume_obj['id'],
                'source_type': 'volume',
            }
            kwargs['block_device_mapping_v2'].append(block_mapping)
        return kwargs

    def wait_for_server(
        self,
        server,
        auto_ip=True,
        ips=None,
        ip_pool=None,
        reuse=True,
        timeout=180,
        nat_destination=None,
    ):
        """
        Wait for a server to reach ACTIVE status.
        """
        server_id = server['id']
        timeout_message = "Timeout waiting for the server to come up."
        start_time = time.time()

        for count in utils.iterate_timeout(
            timeout,
            timeout_message,
            wait=min(5, timeout),
        ):
            try:
                server = self.get_server(server_id)
            except Exception:
                continue
            if not server:
                continue

            # We have more work to do, but the details of that are
            # hidden from the user. So, calculate remaining timeout
            # and pass it down into the IP stack.
            remaining_timeout = timeout - int(time.time() - start_time)
            if remaining_timeout <= 0:
                raise exceptions.ResourceTimeout(timeout_message)

            server = self.get_active_server(
                server=server,
                reuse=reuse,
                auto_ip=auto_ip,
                ips=ips,
                ip_pool=ip_pool,
                wait=True,
                timeout=remaining_timeout,
                nat_destination=nat_destination,
            )

            if server is not None and server['status'] == 'ACTIVE':
                return server

    def get_active_server(
        self,
        server,
        auto_ip=True,
        ips=None,
        ip_pool=None,
        reuse=True,
        wait=False,
        timeout=180,
        nat_destination=None,
    ):
        if server['status'] == 'ERROR':
            if (
                'fault' in server
                and server['fault'] is not None
                and 'message' in server['fault']
            ):
                raise exceptions.SDKException(
                    "Error in creating the server."
                    " Compute service reports fault: {reason}".format(
                        reason=server['fault']['message']
                    ),
                    extra_data=dict(server=server),
                )

            raise exceptions.SDKException(
                "Error in creating the server"
                " (no further information available)",
                extra_data=dict(server=server),
            )

        if server['status'] == 'ACTIVE':
            if 'addresses' in server and server['addresses']:
                return self.add_ips_to_server(
                    server,
                    auto_ip,
                    ips,
                    ip_pool,
                    reuse=reuse,
                    nat_destination=nat_destination,
                    wait=wait,
                    timeout=timeout,
                )

            self.log.debug(
                'Server %(server)s reached ACTIVE state without'
                ' being allocated an IP address.'
                ' Deleting server.',
                {'server': server['id']},
            )
            try:
                self._delete_server(server=server, wait=wait, timeout=timeout)
            except Exception as e:
                raise exceptions.SDKException(
                    'Server reached ACTIVE state without being'
                    ' allocated an IP address AND then could not'
                    ' be deleted: {0}'.format(e),
                    extra_data=dict(server=server),
                )
            raise exceptions.SDKException(
                'Server reached ACTIVE state without being'
                ' allocated an IP address.',
                extra_data=dict(server=server),
            )
        return None

    def rebuild_server(
        self,
        server_id,
        image_id,
        admin_pass=None,
        detailed=False,
        bare=False,
        wait=False,
        timeout=180,
    ):
        """Rebuild a server.

        :param server_id:
        :param image_id:
        :param admin_pass:
        :param detailed:
        :param bare:
        :param wait:
        :param timeout:
        :returns: A compute ``Server`` object.
        """
        kwargs = {}
        if image_id:
            kwargs['image'] = image_id
        if admin_pass:
            kwargs['admin_password'] = admin_pass

        server = self.compute.rebuild_server(server_id, **kwargs)
        if not wait:
            return self._expand_server(server, bare=bare, detailed=detailed)

        admin_pass = server.get('adminPass') or admin_pass
        server = self.compute.wait_for_server(server, wait=timeout)
        if server['status'] == 'ACTIVE':
            server.adminPass = admin_pass

        return self._expand_server(server, detailed=detailed, bare=bare)

    def set_server_metadata(self, name_or_id, metadata):
        """Set metadata in a server instance.

        :param str name_or_id: The name or ID of the server instance to update.
        :param dict metadata: A dictionary with the key=value pairs
            to set in the server instance. It only updates the key=value pairs
            provided. Existing ones will remain untouched.

        :returns: None
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        server = self.get_server(name_or_id, bare=True)
        if not server:
            raise exceptions.SDKException(
                'Invalid Server {server}'.format(server=name_or_id)
            )

        self.compute.set_server_metadata(server=server.id, **metadata)

    def delete_server_metadata(self, name_or_id, metadata_keys):
        """Delete metadata from a server instance.

        :param str name_or_id: The name or ID of the server instance
            to update.
        :param metadata_keys: A list with the keys to be deleted
            from the server instance.

        :returns: None
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        server = self.get_server(name_or_id, bare=True)
        if not server:
            raise exceptions.SDKException(
                'Invalid Server {server}'.format(server=name_or_id)
            )

        self.compute.delete_server_metadata(
            server=server.id, keys=metadata_keys
        )

    def delete_server(
        self,
        name_or_id,
        wait=False,
        timeout=180,
        delete_ips=False,
        delete_ip_retry=1,
    ):
        """Delete a server instance.

        :param name_or_id: name or ID of the server to delete
        :param bool wait: If true, waits for server to be deleted.
        :param int timeout: Seconds to wait for server deletion.
        :param bool delete_ips: If true, deletes any floating IPs
            associated with the instance.
        :param int delete_ip_retry: Number of times to retry deleting
            any floating ips, should the first try be unsuccessful.

        :returns: True if delete succeeded, False otherwise if the
            server does not exist.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        # If delete_ips is True, we need the server to not be bare.
        server = self.compute.find_server(name_or_id, ignore_missing=True)
        if not server:
            return False

        # This portion of the code is intentionally left as a separate
        # private method in order to avoid an unnecessary API call to get
        # a server we already have.
        return self._delete_server(
            server,
            wait=wait,
            timeout=timeout,
            delete_ips=delete_ips,
            delete_ip_retry=delete_ip_retry,
        )

    def _delete_server_floating_ips(self, server, delete_ip_retry):
        # Does the server have floating ips in its
        # addresses dict? If not, skip this.
        server_floats = meta.find_nova_interfaces(
            server['addresses'], ext_tag='floating'
        )
        for fip in server_floats:
            try:
                ip = self.get_floating_ip(
                    id=None, filters={'floating_ip_address': fip['addr']}
                )
            except exceptions.NotFoundException:
                # We're deleting. If it doesn't exist - awesome
                # NOTE(mordred) If the cloud is a nova FIP cloud but
                #               floating_ip_source is set to neutron, this
                #               can lead to a FIP leak.
                continue
            if not ip:
                continue
            deleted = self.delete_floating_ip(ip['id'], retry=delete_ip_retry)
            if not deleted:
                raise exceptions.SDKException(
                    "Tried to delete floating ip {floating_ip}"
                    " associated with server {id} but there was"
                    " an error deleting it. Not deleting server.".format(
                        floating_ip=ip['floating_ip_address'], id=server['id']
                    )
                )

    def _delete_server(
        self,
        server,
        wait=False,
        timeout=180,
        delete_ips=False,
        delete_ip_retry=1,
    ):
        if not server:
            return False

        if delete_ips and self._has_floating_ips() and server['addresses']:
            self._delete_server_floating_ips(server, delete_ip_retry)

        try:
            self.compute.delete_server(server)
        except exceptions.ResourceNotFound:
            return False
        except Exception:
            raise

        if not wait:
            return True

        if not isinstance(server, _server.Server):
            # We might come here with Munch object (at the moment).
            # If this is the case - convert it into real server to be able to
            # use wait_for_delete
            server = _server.Server(id=server['id'])
        self.compute.wait_for_delete(server, wait=timeout)

        return True

    @_utils.valid_kwargs('name', 'description')
    def update_server(self, name_or_id, detailed=False, bare=False, **kwargs):
        """Update a server.

        :param name_or_id: Name of the server to be updated.
        :param detailed: Whether or not to add detailed additional information.
            Defaults to False.
        :param bare: Whether to skip adding any additional information to the
            server record. Defaults to False, meaning the addresses dict will
            be populated as needed from neutron. Setting to True implies
            detailed = False.
        :param name: New name for the server
        :param description: New description for the server

        :returns: The updated compute ``Server`` object.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        server = self.compute.find_server(name_or_id, ignore_missing=False)

        server = self.compute.update_server(server, **kwargs)

        return self._expand_server(server, bare=bare, detailed=detailed)

    def create_server_group(self, name, policies=None, policy=None):
        """Create a new server group.

        :param name: Name of the server group being created
        :param policies: List of policies for the server group.

        :returns: The created compute ``ServerGroup`` object.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        sg_attrs = {'name': name}
        if policies:
            sg_attrs['policies'] = policies
        if policy:
            sg_attrs['policy'] = policy
        return self.compute.create_server_group(**sg_attrs)

    def delete_server_group(self, name_or_id):
        """Delete a server group.

        :param name_or_id: Name or ID of the server group to delete

        :returns: True if delete succeeded, False otherwise
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        server_group = self.get_server_group(name_or_id)
        if not server_group:
            self.log.debug(
                "Server group %s not found for deleting", name_or_id
            )
            return False

        self.compute.delete_server_group(server_group, ignore_missing=False)
        return True

    def create_flavor(
        self,
        name,
        ram,
        vcpus,
        disk,
        description=None,
        flavorid="auto",
        ephemeral=0,
        swap=0,
        rxtx_factor=1.0,
        is_public=True,
    ):
        """Create a new flavor.

        :param name: Descriptive name of the flavor
        :param ram: Memory in MB for the flavor
        :param vcpus: Number of VCPUs for the flavor
        :param disk: Size of local disk in GB
        :param description: Description of the flavor
        :param flavorid: ID for the flavor (optional)
        :param ephemeral: Ephemeral space size in GB
        :param swap: Swap space in MB
        :param rxtx_factor: RX/TX factor
        :param is_public: Make flavor accessible to the public

        :returns: The created compute ``Flavor`` object.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        attrs = {
            'disk': disk,
            'ephemeral': ephemeral,
            'id': flavorid,
            'is_public': is_public,
            'name': name,
            'ram': ram,
            'rxtx_factor': rxtx_factor,
            'swap': swap,
            'vcpus': vcpus,
            'description': description,
        }
        if flavorid == 'auto':
            attrs['id'] = None

        return self.compute.create_flavor(**attrs)

    def delete_flavor(self, name_or_id):
        """Delete a flavor

        :param name_or_id: ID or name of the flavor to delete.

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        try:
            flavor = self.compute.find_flavor(name_or_id)
            if not flavor:
                self.log.debug("Flavor %s not found for deleting", name_or_id)
                return False
            self.compute.delete_flavor(flavor)
            return True
        except exceptions.SDKException:
            raise exceptions.SDKException(
                "Unable to delete flavor {name}".format(name=name_or_id)
            )

    def set_flavor_specs(self, flavor_id, extra_specs):
        """Add extra specs to a flavor

        :param string flavor_id: ID of the flavor to update.
        :param dict extra_specs: Dictionary of key-value pairs.

        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        :raises: :class:`~openstack.exceptions.BadRequestException` if flavor
            ID is not found.
        """
        self.compute.create_flavor_extra_specs(flavor_id, extra_specs)

    def unset_flavor_specs(self, flavor_id, keys):
        """Delete extra specs from a flavor

        :param string flavor_id: ID of the flavor to update.
        :param keys: List of spec keys to delete.

        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        :raises: :class:`~openstack.exceptions.BadRequestException` if flavor
            ID is not found.
        """
        for key in keys:
            self.compute.delete_flavor_extra_specs_property(flavor_id, key)

    def add_flavor_access(self, flavor_id, project_id):
        """Grant access to a private flavor for a project/tenant.

        :param string flavor_id: ID of the private flavor.
        :param string project_id: ID of the project/tenant.

        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        self.compute.flavor_add_tenant_access(flavor_id, project_id)

    def remove_flavor_access(self, flavor_id, project_id):
        """Revoke access from a private flavor for a project/tenant.

        :param string flavor_id: ID of the private flavor.
        :param string project_id: ID of the project/tenant.

        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        self.compute.flavor_remove_tenant_access(flavor_id, project_id)

    def list_flavor_access(self, flavor_id):
        """List access from a private flavor for a project/tenant.

        :param string flavor_id: ID of the private flavor.

        :returns: List of dicts with flavor_id and tenant_id attributes.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        return self.compute.get_flavor_access(flavor_id)

    def list_hypervisors(self, filters=None):
        """List all hypervisors

        :param filters:
        :returns: A list of compute ``Hypervisor`` objects.
        """
        if not filters:
            filters = {}

        return list(self.compute.hypervisors(details=True, **filters))

    def search_aggregates(self, name_or_id=None, filters=None):
        """Seach host aggregates.

        :param name: aggregate name or id.
        :param filters: a dict containing additional filters to use.

        :returns: A list of compute ``Aggregate`` objects matching the search
            criteria.
        :raises: :class:`~openstack.exceptions.SDKException` if something goes
            wrong during the OpenStack API call.
        """
        aggregates = self.list_aggregates()
        return _utils._filter_list(aggregates, name_or_id, filters)

    def list_aggregates(self, filters=None):
        """List all available host aggregates.

        :returns: A list of compute ``Aggregate`` objects.
        """
        if not filters:
            filters = {}

        return self.compute.aggregates(**filters)

    def get_aggregate(self, name_or_id, filters=None):
        """Get an aggregate by name or ID.

        :param name_or_id: Name or ID of the aggregate.
        :param dict filters:
            A dictionary of meta data to use for further filtering. Elements
            of this dictionary may, themselves, be dictionaries. Example::

                {
                    'availability_zone': 'nova',
                    'metadata': {
                        'cpu_allocation_ratio': '1.0'
                    }
                }

        :returns: An aggregate dict or None if no matching aggregate is
            found.
        """
        return self.compute.find_aggregate(name_or_id, ignore_missing=True)

    def create_aggregate(self, name, availability_zone=None):
        """Create a new host aggregate.

        :param name: Name of the host aggregate being created
        :param availability_zone: Availability zone to assign hosts

        :returns: The created compute ``Aggregate`` object.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        return self.compute.create_aggregate(
            name=name, availability_zone=availability_zone
        )

    @_utils.valid_kwargs('name', 'availability_zone')
    def update_aggregate(self, name_or_id, **kwargs):
        """Update a host aggregate.

        :param name_or_id: Name or ID of the aggregate being updated.
        :param name: New aggregate name
        :param availability_zone: Availability zone to assign to hosts

        :returns: The updated compute ``Aggregate`` object.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        aggregate = self.get_aggregate(name_or_id)
        return self.compute.update_aggregate(aggregate, **kwargs)

    def delete_aggregate(self, name_or_id):
        """Delete a host aggregate.

        :param name_or_id: Name or ID of the host aggregate to delete.

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        if isinstance(name_or_id, (str, bytes)) and not name_or_id.isdigit():
            aggregate = self.get_aggregate(name_or_id)
            if not aggregate:
                self.log.debug(
                    "Aggregate %s not found for deleting", name_or_id
                )
                return False
            name_or_id = aggregate.id
        try:
            self.compute.delete_aggregate(name_or_id, ignore_missing=False)
            return True
        except exceptions.ResourceNotFound:
            self.log.debug("Aggregate %s not found for deleting", name_or_id)
            return False

    def set_aggregate_metadata(self, name_or_id, metadata):
        """Set aggregate metadata, replacing the existing metadata.

        :param name_or_id: Name of the host aggregate to update
        :param metadata: Dict containing metadata to replace (Use
            {'key': None} to remove a key)

        :returns: a dict representing the new host aggregate.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        aggregate = self.get_aggregate(name_or_id)
        if not aggregate:
            raise exceptions.SDKException(
                "Host aggregate %s not found." % name_or_id
            )

        return self.compute.set_aggregate_metadata(aggregate, metadata)

    def add_host_to_aggregate(self, name_or_id, host_name):
        """Add a host to an aggregate.

        :param name_or_id: Name or ID of the host aggregate.
        :param host_name: Host to add.

        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        aggregate = self.get_aggregate(name_or_id)
        if not aggregate:
            raise exceptions.SDKException(
                "Host aggregate %s not found." % name_or_id
            )

        return self.compute.add_host_to_aggregate(aggregate, host_name)

    def remove_host_from_aggregate(self, name_or_id, host_name):
        """Remove a host from an aggregate.

        :param name_or_id: Name or ID of the host aggregate.
        :param host_name: Host to remove.

        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        aggregate = self.get_aggregate(name_or_id)
        if not aggregate:
            raise exceptions.SDKException(
                "Host aggregate %s not found." % name_or_id
            )

        return self.compute.remove_host_from_aggregate(aggregate, host_name)

    def set_compute_quotas(self, name_or_id, **kwargs):
        """Set a quota in a project

        :param name_or_id: project name or id
        :param kwargs: key/value pairs of quota name and quota value

        :raises: :class:`~openstack.exceptions.SDKException` if the resource to
            set the quota does not exist.
        """
        proj = self.identity.find_project(name_or_id, ignore_missing=False)
        kwargs['force'] = True
        self.compute.update_quota_set(
            _qs.QuotaSet(project_id=proj.id), **kwargs
        )

    def get_compute_quotas(self, name_or_id):
        """Get quota for a project

        :param name_or_id: project name or id

        :returns: A compute ``QuotaSet`` object if found, else None.
        :raises: :class:`~openstack.exceptions.SDKException` if it's not a
            valid project
        """
        proj = self.identity.find_project(name_or_id, ignore_missing=False)
        return self.compute.get_quota_set(proj)

    def delete_compute_quotas(self, name_or_id):
        """Delete quota for a project

        :param name_or_id: project name or id

        :raises: :class:`~openstack.exceptions.SDKException` if it's not a
            valid project or the nova client call failed
        """
        proj = self.identity.find_project(name_or_id, ignore_missing=False)
        self.compute.revert_quota_set(proj)

    def get_compute_usage(self, name_or_id, start=None, end=None):
        """Get usage for a specific project

        :param name_or_id: project name or id
        :param start: :class:`datetime.datetime` or string. Start date in UTC
            Defaults to 2010-07-06T12:00:00Z (the date the OpenStack project
            was started)
        :param end: :class:`datetime.datetime` or string. End date in UTC.
            Defaults to now

        :returns: A :class:`~openstack.compute.v2.usage.Usage` object
        :raises: :class:`~openstack.exceptions.SDKException` if it's not a
            valid project
        """

        def parse_date(date):
            try:
                return iso8601.parse_date(date)
            except iso8601.iso8601.ParseError:
                # Yes. This is an exception mask. However,iso8601 is an
                # implementation detail - and the error message is actually
                # less informative.
                raise exceptions.SDKException(
                    "Date given, {date}, is invalid. Please pass in a date"
                    " string in ISO 8601 format -"
                    " YYYY-MM-DDTHH:MM:SS".format(date=date)
                )

        if isinstance(start, str):
            start = parse_date(start)
        if isinstance(end, str):
            end = parse_date(end)

        proj = self.get_project(name_or_id)
        if not proj:
            raise exceptions.SDKException(
                "project does not exist: {name}".format(name=proj.id)
            )

        return self.compute.get_usage(proj, start, end)

    def _encode_server_userdata(self, userdata):
        if hasattr(userdata, 'read'):
            userdata = userdata.read()

        if not isinstance(userdata, bytes):
            # If the userdata passed in is bytes, just send it unmodified
            if not isinstance(userdata, str):
                raise TypeError("%s can't be encoded" % type(userdata))
            # If it's not bytes, make it bytes
            userdata = userdata.encode('utf-8', 'strict')

        # Once we have base64 bytes, make them into a utf-8 string for REST
        return base64.b64encode(userdata).decode('utf-8')

    def get_openstack_vars(self, server):
        return meta.get_hostvars_from_server(self, server)

    def _expand_server_vars(self, server):
        # Used by nodepool
        # TODO(mordred) remove after these make it into what we
        # actually want the API to be.
        return meta.expand_server_vars(self, server)

    def _remove_novaclient_artifacts(self, item):
        # Remove novaclient artifacts
        item.pop('links', None)
        item.pop('NAME_ATTR', None)
        item.pop('HUMAN_ID', None)
        item.pop('human_id', None)
        item.pop('request_ids', None)
        item.pop('x_openstack_request_ids', None)

    def _normalize_server(self, server):
        ret = utils.Munch()
        # Copy incoming server because of shared dicts in unittests
        # Wrap the copy in munch so that sub-dicts are properly munched
        server = utils.Munch(server)

        self._remove_novaclient_artifacts(server)

        ret['id'] = server.pop('id')
        ret['name'] = server.pop('name')

        server['flavor'].pop('links', None)
        ret['flavor'] = server.pop('flavor')
        # From original_names from sdk
        server.pop('flavorRef', None)

        # OpenStack can return image as a string when you've booted
        # from volume
        image = server.pop('image', None)
        if str(image) != image:
            image = utils.Munch(id=image['id'])

        ret['image'] = image
        # From original_names from sdk
        server.pop('imageRef', None)
        # From original_names from sdk
        ret['block_device_mapping'] = server.pop('block_device_mapping_v2', {})

        project_id = server.pop('tenant_id', '')
        project_id = server.pop('project_id', project_id)

        az = _pop_or_get(
            server, 'OS-EXT-AZ:availability_zone', None, self.strict_mode
        )
        # the server resource has this already, but it's missing az info
        # from the resource.
        # TODO(mordred) create_server is still normalizing servers that aren't
        # from the resource layer.
        ret['location'] = server.pop(
            'location',
            self._get_current_location(project_id=project_id, zone=az),
        )

        # Ensure volumes is always in the server dict, even if empty
        ret['volumes'] = _pop_or_get(
            server,
            'os-extended-volumes:volumes_attached',
            [],
            self.strict_mode,
        )

        config_drive = server.pop(
            'has_config_drive', server.pop('config_drive', False)
        )
        ret['has_config_drive'] = _to_bool(config_drive)

        host_id = server.pop('hostId', server.pop('host_id', None))
        ret['host_id'] = host_id

        ret['progress'] = _pop_int(server, 'progress')

        # Leave these in so that the general properties handling works
        ret['disk_config'] = _pop_or_get(
            server, 'OS-DCF:diskConfig', None, self.strict_mode
        )
        for key in (
            'OS-EXT-STS:power_state',
            'OS-EXT-STS:task_state',
            'OS-EXT-STS:vm_state',
            'OS-SRV-USG:launched_at',
            'OS-SRV-USG:terminated_at',
            'OS-EXT-SRV-ATTR:hypervisor_hostname',
            'OS-EXT-SRV-ATTR:instance_name',
            'OS-EXT-SRV-ATTR:user_data',
            'OS-EXT-SRV-ATTR:host',
            'OS-EXT-SRV-ATTR:hostname',
            'OS-EXT-SRV-ATTR:kernel_id',
            'OS-EXT-SRV-ATTR:launch_index',
            'OS-EXT-SRV-ATTR:ramdisk_id',
            'OS-EXT-SRV-ATTR:reservation_id',
            'OS-EXT-SRV-ATTR:root_device_name',
            'OS-SCH-HNT:scheduler_hints',
        ):
            short_key = key.split(':')[1]
            ret[short_key] = _pop_or_get(server, key, None, self.strict_mode)

        # Protect against security_groups being None
        ret['security_groups'] = server.pop('security_groups', None) or []

        # NOTE(mnaser): The Nova API returns the creation date in `created`
        #               however the Shade contract returns `created_at` for
        #               all resources.
        ret['created_at'] = server.get('created')

        for field in _SERVER_FIELDS:
            ret[field] = server.pop(field, None)
        if not ret['networks']:
            ret['networks'] = {}

        ret['interface_ip'] = ''

        ret['properties'] = server.copy()

        # Backwards compat
        if not self.strict_mode:
            ret['hostId'] = host_id
            ret['config_drive'] = config_drive
            ret['project_id'] = project_id
            ret['tenant_id'] = project_id
            # TODO(efried): This is hardcoded to 'compute' because this method
            # should only ever be used by the compute proxy. (That said, it
            # doesn't appear to be used at all, so can we get rid of it?)
            ret['region'] = self.config.get_region_name('compute')
            ret['cloud'] = self.config.name
            ret['az'] = az
            for key, val in ret['properties'].items():
                ret.setdefault(key, val)
        return ret



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack import resource
from openstack import utils


class ServerGroup(resource.Resource):
    resource_key = 'server_group'
    resources_key = 'server_groups'
    base_path = '/os-server-groups'

    _query_mapping = resource.QueryParameters("all_projects")

    _max_microversion = '2.64'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_delete = True
    allow_list = True

    # Properties
    #: A name identifying the server group
    name = resource.Body('name')
    #: The list of policies supported by the server group (till 2.63)
    policies = resource.Body('policies')
    #: The policy field represents the name of the policy (from 2.64)
    policy = resource.Body('policy')
    #: The list of members in the server group
    member_ids = resource.Body('members')
    #: The metadata associated with the server group. This is always empty and
    #: only used for preserving compatibility.
    metadata = resource.Body('metadata')
    #: The project ID who owns the server group.
    project_id = resource.Body('project_id')
    #: The rules field, which is a dict, can be applied to the policy.
    #: Currently, only the max_server_per_host rule is supported for the
    #: anti-affinity policy. The max_server_per_host rule allows specifying how
    #: many members of the anti-affinity group can reside on the same compute
    #: host. If not specified, only one member from the same anti-affinity
    #: group can reside on a given host.
    rules = resource.Body('rules', type=dict)
    #: The user ID who owns the server group
    user_id = resource.Body('user_id')

    # TODO(stephenfin): It would be nice to have a hookpoint to do this
    # microversion-based request manipulation, but we don't have anything like
    # that right now
    def create(self, session, prepend_key=True, base_path=None, **params):
        """Create a remote resource based on this instance.

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param prepend_key: A boolean indicating whether the resource_key
            should be prepended in a resource creation request. Default to
            True.
        :param str base_path: Base part of the URI for creating resources, if
            different from :data:`~openstack.resource.Resource.base_path`.
        :param dict params: Additional params to pass.
        :return: This :class:`Resource` instance.
        :raises: :exc:`~openstack.exceptions.MethodNotSupported` if
            :data:`Resource.allow_create` is not set to ``True``.
        """
        if not self.allow_create:
            raise exceptions.MethodNotSupported(self, 'create')

        session = self._get_session(session)
        microversion = self._get_microversion(session, action='create')
        requires_id = (
            self.create_requires_id
            if self.create_requires_id is not None
            else self.create_method == 'PUT'
        )

        if self.create_exclude_id_from_body:
            self._body._dirty.discard("id")

        # `policy` and `rules` are added with mv=2.64. In it also
        # `policies` are removed.
        if utils.supports_microversion(session, '2.64'):
            if self.policies:
                if not self.policy and isinstance(self.policies, list):
                    self.policy = self.policies[0]
                self._body.clean(only={'policies'})
            microversion = self._max_microversion
        else:  # microversion < 2.64
            if self.rules:
                msg = (
                    "API version 2.64 is required to set rules, but "
                    "it is not available."
                )
                raise exceptions.NotSupported(msg)

            if self.policy:
                if not self.policies:
                    self.policies = [self.policy]
                self._body.clean(only={'policy'})

        if self.create_method == 'POST':
            request = self._prepare_request(
                requires_id=requires_id,
                prepend_key=prepend_key,
                base_path=base_path,
            )
            response = session.post(
                request.url,
                json=request.body,
                headers=request.headers,
                microversion=microversion,
                params=params,
            )
        else:
            raise exceptions.ResourceFailure(
                "Invalid create method: %s" % self.create_method
            )

        has_body = (
            self.has_body
            if self.create_returns_body is None
            else self.create_returns_body
        )
        self.microversion = microversion
        self._translate_response(response, has_body=has_body)
        # direct comparision to False since we need to rule out None
        if self.has_body and self.create_returns_body is False:
            # fetch the body if it's required but not returned by create
            return self.fetch(session)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class StackEvent(resource.Resource):
    base_path = '/stacks/%(stack_name)s/%(stack_id)s/events'
    resources_key = 'events'

    # capabilities
    allow_create = False
    allow_list = True
    allow_fetch = True
    allow_delete = False
    allow_commit = False

    _query_mapping = resource.QueryParameters(
        "resource_action",
        "resource_status",
        "resource_name",
        "resource_type",
        "nested_depth",
        "sort_key",
        "sort_dir",
    )

    # Properties
    #: The date and time when the event was created
    event_time = resource.Body('event_time')
    #: The ID of the event object
    id = resource.Body('id')
    #: A list of dictionaries containing links relevant to the stack.
    links = resource.Body('links')
    #: The ID of the logical stack resource.
    logical_resource_id = resource.Body('logical_resource_id')
    #: The ID of the stack physical resource.
    physical_resource_id = resource.Body('physical_resource_id')
    #: The name of the resource.
    resource_name = resource.Body('resource_name')
    #: The status of the resource.
    resource_status = resource.Body('resource_status')
    #: The reason for the current stack resource state.
    resource_status_reason = resource.Body('resource_status_reason')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.shared_file_system.v2 import share
from openstack.tests.unit import base


IDENTIFIER = '08a87d37-5ca2-4308-86c5-cba06d8d796c'
EXAMPLE = {
    "id": IDENTIFIER,
    "size": 2,
    "availability_zone": "manila-zone-1",
    "created_at": "2021-02-11T17:38:00.000000",
    "status": "available",
    "name": None,
    "description": None,
    "project_id": "d19444eb73af4b37bc0794532ef6fc50",
    "snapshot_id": None,
    "share_network_id": None,
    "share_protocol": "NFS",
    "metadata": {},
    "share_type": "cbb18bb7-cc97-477a-b64b-ed7c7f2a1c67",
    "volume_type": "default",
    "is_public": False,
    "is_snapshot_supported": True,
    "task_state": None,
    "share_type_name": "default",
    "access_rules_status": "active",
    "replication_type": None,
    "is_replicated": False,
    "user_id": "6c262cab98de42c2afc4cfccbefc50c7",
    "is_creating_new_share_from_snapshot_supported": True,
    "is_reverting_to_snapshot_supported": True,
    "share_group_id": None,
    "source_share_group_snapshot_member_id": None,
    "is_mounting_snapshot_supported": True,
    "progress": "100%",
    "share_server_id": None,
    "host": "new@denver#lvm-single-pool",
}


class TestShares(base.TestCase):
    def test_basic(self):
        shares_resource = share.Share()
        self.assertEqual('shares', shares_resource.resources_key)
        self.assertEqual('/shares', shares_resource.base_path)
        self.assertTrue(shares_resource.allow_list)
        self.assertTrue(shares_resource.allow_create)
        self.assertTrue(shares_resource.allow_fetch)
        self.assertTrue(shares_resource.allow_commit)
        self.assertTrue(shares_resource.allow_delete)

    def test_make_shares(self):
        shares_resource = share.Share(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], shares_resource.id)
        self.assertEqual(EXAMPLE['size'], shares_resource.size)
        self.assertEqual(
            EXAMPLE['availability_zone'], shares_resource.availability_zone
        )
        self.assertEqual(EXAMPLE['created_at'], shares_resource.created_at)
        self.assertEqual(EXAMPLE['status'], shares_resource.status)
        self.assertEqual(EXAMPLE['name'], shares_resource.name)
        self.assertEqual(EXAMPLE['description'], shares_resource.description)
        self.assertEqual(EXAMPLE['project_id'], shares_resource.project_id)
        self.assertEqual(EXAMPLE['snapshot_id'], shares_resource.snapshot_id)
        self.assertEqual(
            EXAMPLE['share_network_id'], shares_resource.share_network_id
        )
        self.assertEqual(
            EXAMPLE['share_protocol'], shares_resource.share_protocol
        )
        self.assertEqual(EXAMPLE['metadata'], shares_resource.metadata)
        self.assertEqual(EXAMPLE['share_type'], shares_resource.share_type)
        self.assertEqual(EXAMPLE['is_public'], shares_resource.is_public)
        self.assertEqual(
            EXAMPLE['is_snapshot_supported'],
            shares_resource.is_snapshot_supported,
        )
        self.assertEqual(EXAMPLE['task_state'], shares_resource.task_state)
        self.assertEqual(
            EXAMPLE['share_type_name'], shares_resource.share_type_name
        )
        self.assertEqual(
            EXAMPLE['access_rules_status'], shares_resource.access_rules_status
        )
        self.assertEqual(
            EXAMPLE['replication_type'], shares_resource.replication_type
        )
        self.assertEqual(
            EXAMPLE['is_replicated'], shares_resource.is_replicated
        )
        self.assertEqual(EXAMPLE['user_id'], shares_resource.user_id)
        self.assertEqual(
            EXAMPLE['is_creating_new_share_from_snapshot_supported'],
            (shares_resource.is_creating_new_share_from_snapshot_supported),
        )
        self.assertEqual(
            EXAMPLE['is_reverting_to_snapshot_supported'],
            shares_resource.is_reverting_to_snapshot_supported,
        )
        self.assertEqual(
            EXAMPLE['share_group_id'], shares_resource.share_group_id
        )
        self.assertEqual(
            EXAMPLE['source_share_group_snapshot_member_id'],
            shares_resource.source_share_group_snapshot_member_id,
        )
        self.assertEqual(
            EXAMPLE['is_mounting_snapshot_supported'],
            shares_resource.is_mounting_snapshot_supported,
        )
        self.assertEqual(EXAMPLE['progress'], shares_resource.progress)
        self.assertEqual(
            EXAMPLE['share_server_id'], shares_resource.share_server_id
        )
        self.assertEqual(EXAMPLE['host'], shares_resource.host)


class TestShareActions(TestShares):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.status_code = 202
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.default_microversion = '3.0'
        self.sess.post = mock.Mock(return_value=self.resp)
        self.sess._get_connection = mock.Mock(return_value=self.cloud)

    def test_shrink_share(self):
        sot = share.Share(**EXAMPLE)
        microversion = sot._get_microversion(self.sess, action='patch')

        self.assertIsNone(sot.shrink_share(self.sess, new_size=1))

        url = f'shares/{IDENTIFIER}/action'
        body = {"shrink": {"new_size": 1}}
        headers = {'Accept': ''}

        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion=microversion
        )

    def test_extend_share(self):
        sot = share.Share(**EXAMPLE)
        microversion = sot._get_microversion(self.sess, action='patch')

        self.assertIsNone(sot.extend_share(self.sess, new_size=3))

        url = f'shares/{IDENTIFIER}/action'
        body = {"extend": {"new_size": 3}}
        headers = {'Accept': ''}

        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion=microversion
        )

    def test_revert_to_snapshot(self):
        sot = share.Share(**EXAMPLE)
        microversion = sot._get_microversion(self.sess, action='patch')

        self.assertIsNone(sot.revert_to_snapshot(self.sess, "fake_id"))

        url = f'shares/{IDENTIFIER}/action'
        body = {"revert": {"snapshot_id": "fake_id"}}
        headers = {'Accept': ''}

        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion=microversion
        )

    def test_manage_share(self):
        sot = share.Share()

        self.resp.headers = {}
        self.resp.json = mock.Mock(
            return_value={"share": {"name": "test_share", "size": 1}}
        )

        export_path = (
            "10.254.0 .5:/shares/share-42033c24-0261-424f-abda-4fef2f6dbfd5."
        )
        params = {"name": "test_share"}
        res = sot.manage(
            self.sess,
            sot["share_protocol"],
            export_path,
            sot["host"],
            **params,
        )

        self.assertEqual(res.name, "test_share")
        self.assertEqual(res.size, 1)

        jsonDict = {
            "share": {
                "protocol": sot["share_protocol"],
                "export_path": export_path,
                "service_host": sot["host"],
                "name": "test_share",
            }
        }

        self.sess.post.assert_called_once_with("shares/manage", json=jsonDict)

    def test_unmanage_share(self):
        sot = share.Share(**EXAMPLE)
        microversion = sot._get_microversion(self.sess, action='patch')

        self.assertIsNone(sot.unmanage(self.sess))

        url = 'shares/%s/action' % IDENTIFIER
        body = {'unmanage': None}

        self.sess.post.assert_called_with(
            url, json=body, headers={'Accept': ''}, microversion=microversion
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class DomainConfigLDAP(resource.Resource):
    #: The base distinguished name (DN) of LDAP.
    user_tree_dn = resource.Body('user_tree_dn')
    #: The LDAP URL.
    url = resource.Body('url')


class DomainConfigDriver(resource.Resource):
    #: The Identity backend driver.
    driver = resource.Body('driver')


class DomainConfig(resource.Resource):
    resource_key = 'config'
    base_path = '/domains/%(domain_id)s/config'
    requires_id = False
    create_requires_id = False
    commit_method = 'PATCH'
    create_method = 'PUT'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True

    #: The domain ID.
    domain_id = resource.URI('domain_id')
    #: An identity object.
    identity = resource.Body('identity', type=DomainConfigDriver)
    #: The config object.
    ldap = resource.Body('ldap', type=DomainConfigLDAP)



#   Copyright 2021 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from openstack import resource


class LocalIP(resource.Resource):
    """Local IP extension."""

    resource_name = "local ip"
    resource_key = "local_ip"
    resources_key = "local_ips"
    base_path = "/local_ips"

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _allow_unknown_attrs_in_body = True

    _query_mapping = resource.QueryParameters(
        "sort_key",
        "sort_dir",
        'name',
        'description',
        'project_id',
        'network_id',
        'local_port_id',
        'local_ip_address',
        'ip_mode',
    )

    # Properties
    #: Timestamp at which the floating IP was created.
    created_at = resource.Body('created_at')
    #: The local ip description.
    description = resource.Body('description')
    #: The ID of the local ip.
    id = resource.Body('id')
    #: The local ip ip-mode.
    ip_mode = resource.Body('ip_mode')
    #: The Local IP address.
    local_ip_address = resource.Body('local_ip_address')
    #: The ID of the port that owns the local ip.
    local_port_id = resource.Body('local_port_id')
    #: The local ip name.
    name = resource.Body('name')
    #: The ID of the network that owns the local ip.
    network_id = resource.Body('network_id')
    #: The ID of the project that owns the local ip.
    project_id = resource.Body('project_id')
    #: The local ip revision number.
    revision_number = resource.Body('revision_number')
    #: Timestamp at which the floating IP was last updated.
    updated_at = resource.Body('updated_at')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v3 import group as _group
from openstack.block_storage.v3 import group_snapshot as _group_snapshot
from openstack.block_storage.v3 import group_type as _group_type
from openstack.block_storage.v3 import volume as _volume
from openstack.tests.functional.block_storage.v3 import base


class TestGroup(base.BaseBlockStorageTest):
    # TODO(stephenfin): We should use setUpClass here for MOAR SPEED!!!
    def setUp(self):
        super().setUp()

        # there will always be at least one volume type, i.e. the default one
        volume_types = list(self.conn.block_storage.types())
        self.volume_type = volume_types[0]

        group_type_name = self.getUniqueString()
        self.group_type = self.conn.block_storage.create_group_type(
            name=group_type_name,
        )
        self.assertIsInstance(self.group_type, _group_type.GroupType)
        self.assertEqual(group_type_name, self.group_type.name)

        group_name = self.getUniqueString()
        self.group = self.conn.block_storage.create_group(
            name=group_name,
            group_type=self.group_type.id,
            volume_types=[self.volume_type.id],
        )
        self.assertIsInstance(self.group, _group.Group)
        self.assertEqual(group_name, self.group.name)

    def tearDown(self):
        # we do this in tearDown rather than via 'addCleanup' since we need to
        # wait for the deletion of the group before moving onto the deletion of
        # the group type
        self.conn.block_storage.delete_group(self.group, delete_volumes=True)
        self.conn.block_storage.wait_for_delete(self.group)

        self.conn.block_storage.delete_group_type(self.group_type)
        self.conn.block_storage.wait_for_delete(self.group_type)

        super().tearDown()

    def test_group_type(self):
        # get
        group_type = self.conn.block_storage.get_group_type(self.group_type.id)
        self.assertEqual(self.group_type.name, group_type.name)

        # find
        group_type = self.conn.block_storage.find_group_type(
            self.group_type.name,
        )
        self.assertEqual(self.group_type.id, group_type.id)

        # list
        group_types = list(self.conn.block_storage.group_types())
        # other tests may have created group types and there can be defaults so
        # we don't assert that this is the *only* group type present
        self.assertIn(self.group_type.id, {g.id for g in group_types})

        # update
        group_type_name = self.getUniqueString()
        group_type_description = self.getUniqueString()
        group_type = self.conn.block_storage.update_group_type(
            self.group_type,
            name=group_type_name,
            description=group_type_description,
        )
        self.assertIsInstance(group_type, _group_type.GroupType)
        group_type = self.conn.block_storage.get_group_type(self.group_type.id)
        self.assertEqual(group_type_name, group_type.name)
        self.assertEqual(group_type_description, group_type.description)

    def test_group_type_group_specs(self):
        # create
        group_type = self.conn.block_storage.create_group_type_group_specs(
            self.group_type,
            {'foo': 'bar', 'acme': 'buzz'},
        )
        self.assertIsInstance(group_type, _group_type.GroupType)
        group_type = self.conn.block_storage.get_group_type(self.group_type.id)
        self.assertEqual(
            {'foo': 'bar', 'acme': 'buzz'}, group_type.group_specs
        )

        # get
        spec = self.conn.block_storage.get_group_type_group_specs_property(
            self.group_type,
            'foo',
        )
        self.assertEqual('bar', spec)

        # update
        spec = self.conn.block_storage.update_group_type_group_specs_property(
            self.group_type,
            'foo',
            'baz',
        )
        self.assertEqual('baz', spec)
        group_type = self.conn.block_storage.get_group_type(self.group_type.id)
        self.assertEqual(
            {'foo': 'baz', 'acme': 'buzz'}, group_type.group_specs
        )

        # delete
        self.conn.block_storage.delete_group_type_group_specs_property(
            self.group_type,
            'foo',
        )
        group_type = self.conn.block_storage.get_group_type(self.group_type.id)
        self.assertEqual({'acme': 'buzz'}, group_type.group_specs)

    def test_group(self):
        # get
        group = self.conn.block_storage.get_group(self.group.id)
        self.assertEqual(self.group.name, group.name)

        # find
        group = self.conn.block_storage.find_group(self.group.name)
        self.assertEqual(self.group.id, group.id)

        # list
        groups = self.conn.block_storage.groups()
        # other tests may have created groups and there can be defaults so we
        # don't assert that this is the *only* group present
        self.assertIn(self.group.id, {g.id for g in groups})

        # update
        group_name = self.getUniqueString()
        group_description = self.getUniqueString()
        group = self.conn.block_storage.update_group(
            self.group,
            name=group_name,
            description=group_description,
        )
        self.assertIsInstance(group, _group.Group)
        group = self.conn.block_storage.get_group(self.group.id)
        self.assertEqual(group_name, group.name)
        self.assertEqual(group_description, group.description)

    def test_group_snapshot(self):
        # group snapshots require a volume
        # no need for a teardown as the deletion of the group (with the
        # 'delete_volumes' flag) will handle this but we do need to wait for
        # the thing to be created
        volume_name = self.getUniqueString()
        self.volume = self.conn.block_storage.create_volume(
            name=volume_name,
            volume_type=self.volume_type.id,
            group_id=self.group.id,
            size=1,
        )
        self.conn.block_storage.wait_for_status(
            self.volume,
            status='available',
            failures=['error'],
            interval=2,
            wait=self._wait_for_timeout,
        )
        self.assertIsInstance(self.volume, _volume.Volume)

        group_snapshot_name = self.getUniqueString()
        self.group_snapshot = self.conn.block_storage.create_group_snapshot(
            name=group_snapshot_name,
            group_id=self.group.id,
        )
        self.conn.block_storage.wait_for_status(
            self.group_snapshot,
            status='available',
            failures=['error'],
            interval=2,
            wait=self._wait_for_timeout,
        )
        self.assertIsInstance(
            self.group_snapshot,
            _group_snapshot.GroupSnapshot,
        )

        # get
        group_snapshot = self.conn.block_storage.get_group_snapshot(
            self.group_snapshot.id,
        )
        self.assertEqual(self.group_snapshot.name, group_snapshot.name)

        # find
        group_snapshot = self.conn.block_storage.find_group_snapshot(
            self.group_snapshot.name,
        )
        self.assertEqual(self.group_snapshot.id, group_snapshot.id)

        # list
        group_snapshots = self.conn.block_storage.group_snapshots()
        # other tests may have created group snapshot and there can be defaults
        # so we don't assert that this is the *only* group snapshot present
        self.assertIn(self.group_snapshot.id, {g.id for g in group_snapshots})

        # update (not supported)

        # delete
        self.conn.block_storage.delete_group_snapshot(self.group_snapshot)
        self.conn.block_storage.wait_for_delete(self.group_snapshot)



# Copyright (c) 2018 China Telecom Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstack import resource


class Quota(resource.Resource):
    resource_key = 'quota'
    resources_key = 'quotas'
    base_path = '/lbaas/quotas'

    # capabilities
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Properties
    #: The maximum amount of load balancers you can have. *Type: int*
    load_balancers = resource.Body('load_balancer', type=int)
    #: The maximum amount of listeners you can create. *Type: int*
    listeners = resource.Body('listener', type=int)
    #: The maximum amount of pools you can create. *Type: int*
    pools = resource.Body('pool', type=int)
    #: The maximum amount of health monitors you can create. *Type: int*
    health_monitors = resource.Body('health_monitor', type=int)
    #: The maximum amount of members you can create. *Type: int*
    members = resource.Body('member', type=int)
    #: The ID of the project this quota is associated with.
    project_id = resource.Body('project_id', alternate_id=True)

    def _prepare_request(
        self, requires_id=True, base_path=None, prepend_key=False, **kwargs
    ):
        _request = super(Quota, self)._prepare_request(
            requires_id, prepend_key, base_path=base_path
        )
        if self.resource_key in _request.body:
            _body = _request.body[self.resource_key]
        else:
            _body = _request.body
        if 'id' in _body:
            del _body['id']
        return _request


class QuotaDefault(Quota):
    base_path = '/lbaas/quotas/defaults'

    allow_retrieve = True
    allow_commit = False
    allow_delete = False
    allow_list = False



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import application_credential
from openstack.tests.unit import base


EXAMPLE = {
    "user": {"id": "8ac43bb0926245cead88676a96c750d3"},
    "name": 'monitoring',
    "secret": 'rEaqvJka48mpv',
    "roles": [{"name": "Reader"}],
    "expires_at": '2018-02-27T18:30:59Z',
    "description": "Application credential for monitoring",
    "unrestricted": "False",
    "project_id": "3",
    "links": {"self": "http://example.com/v3/application_credential_1"},
}


class TestApplicationCredential(base.TestCase):
    def test_basic(self):
        sot = application_credential.ApplicationCredential()
        self.assertEqual('application_credential', sot.resource_key)
        self.assertEqual('application_credentials', sot.resources_key)
        self.assertEqual(
            '/users/%(user_id)s/application_credentials', sot.base_path
        )
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = application_credential.ApplicationCredential(**EXAMPLE)
        self.assertEqual(EXAMPLE['user'], sot.user)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['secret'], sot.secret)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['expires_at'], sot.expires_at)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['roles'], sot.roles)
        self.assertEqual(EXAMPLE['links'], sot.links)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
The :mod:`~openstack.test.fakes` module exists to help application developers
using the OpenStack SDK to unit test their applications. It provides a number
of helper utilities to generate fake :class:`~openstack.resource.Resource` and
:class:`~openstack.proxy.Proxy` instances. These fakes do not require an
established connection and allow you to validate that your application using
valid attributes and methods for both :class:`~openstack.resource.Resource` and
:class:`~openstack.proxy.Proxy` instances.
"""

import inspect
import random
from typing import (
    Any,
    Dict,
    Generator,
    Optional,
    Type,
    TypeVar,
)
from unittest import mock
import uuid

from openstack import format as _format
from openstack import proxy
from openstack import resource
from openstack import service_description


Resource = TypeVar('Resource', bound=resource.Resource)


def generate_fake_resource(
    resource_type: Type[Resource],
    **attrs: Dict[str, Any],
) -> Resource:
    """Generate a fake resource

    :param type resource_type: Object class
    :param dict attrs: Optional attributes to be set on resource

    Example usage:

    .. code-block:: python

        >>> from openstack.compute.v2 import server
        >>> from openstack.test import fakes
        >>> fakes.generate_fake_resource(server.Server)
        openstack.compute.v2.server.Server(...)

    :param type resource_type: Object class
    :param dict attrs: Optional attributes to be set on resource
    :return: Instance of ``resource_type`` class populated with fake
        values of expected types
    :raises NotImplementedError: If a resource attribute specifies a ``type``
        or ``list_type`` that cannot be automatically generated
    """
    base_attrs: Dict[str, Any] = {}
    for name, value in inspect.getmembers(
        resource_type,
        predicate=lambda x: isinstance(x, (resource.Body, resource.URI)),
    ):
        if isinstance(value, resource.Body):
            target_type = value.type
            if target_type is None:
                if (
                    name == "properties"
                    and hasattr(
                        resource_type, "_store_unknown_attrs_as_properties"
                    )
                    and resource_type._store_unknown_attrs_as_properties
                ):
                    # virtual "properties" attr which hosts all unknown attrs
                    # (i.e. Image)
                    base_attrs[name] = dict()
                else:
                    # Type not defined - string
                    base_attrs[name] = uuid.uuid4().hex
            elif issubclass(target_type, resource.Resource):
                # Attribute is of another Resource type
                base_attrs[name] = generate_fake_resource(target_type)
            elif issubclass(target_type, list) and value.list_type is not None:
                # List of ...
                item_type = value.list_type
                if issubclass(item_type, resource.Resource):
                    # item is of Resource type
                    base_attrs[name] = [generate_fake_resource(item_type)]
                elif issubclass(item_type, dict):
                    base_attrs[name] = [{}]
                elif issubclass(item_type, str):
                    base_attrs[name] = [uuid.uuid4().hex]
                else:
                    # Everything else
                    msg = "Fake value for %s.%s can not be generated" % (
                        resource_type.__name__,
                        name,
                    )
                    raise NotImplementedError(msg)
            elif issubclass(target_type, list) and value.list_type is None:
                # List of str
                base_attrs[name] = [uuid.uuid4().hex]
            elif issubclass(target_type, str):
                # definitely string
                base_attrs[name] = uuid.uuid4().hex
            elif issubclass(target_type, int):
                # int
                base_attrs[name] = random.randint(1, 100)
            elif issubclass(target_type, float):
                # float
                base_attrs[name] = random.random()
            elif issubclass(target_type, bool) or issubclass(
                target_type, _format.BoolStr
            ):
                # bool
                base_attrs[name] = random.choice([True, False])
            elif issubclass(target_type, dict):
                # some dict - without further details leave it empty
                base_attrs[name] = dict()
            else:
                # Everything else
                msg = "Fake value for %s.%s can not be generated" % (
                    resource_type.__name__,
                    name,
                )
                raise NotImplementedError(msg)

        if isinstance(value, resource.URI):
            # For URI we just generate something
            base_attrs[name] = uuid.uuid4().hex

    base_attrs.update(**attrs)
    fake = resource_type(**base_attrs)
    return fake


def generate_fake_resources(
    resource_type: Type[Resource],
    count: int = 1,
    attrs: Optional[Dict[str, Any]] = None,
) -> Generator[Resource, None, None]:
    """Generate a given number of fake resource entities

    :param type resource_type: Object class
    :param int count: Number of objects to return
    :param dict attrs: Attribute values to set into each instance

    Example usage:

    .. code-block:: python

        >>> from openstack.compute.v2 import server
        >>> from openstack.test import fakes
        >>> fakes.generate_fake_resources(server.Server, count=3)
        <generator object generate_fake_resources at 0x7f075dc65040>

    :param type resource_type: Object class
    :param int count: Number of objects to return
    :param dict attrs: Attribute values to set into each instance
    :return: Generator of ``resource_type`` class instances populated with fake
        values of expected types.
    """
    if not attrs:
        attrs = {}
    for _ in range(count):
        yield generate_fake_resource(resource_type, **attrs)


# TODO(stephenfin): It would be helpful to generate fake resources for the
# various proxy methods also, but doing so requires deep code introspection or
# (better) type annotations
def generate_fake_proxy(
    service: Type[service_description.ServiceDescription],
    api_version: Optional[str] = None,
) -> proxy.Proxy:
    """Generate a fake proxy for the given service type

    Example usage:

    .. code-block:: python

        >>> import functools
        >>> from openstack.compute import compute_service
        >>> from openstack.compute.v2 import server
        >>> from openstack.test import fakes
        >>> # create the fake proxy
        >>> fake_compute_proxy = fakes.generate_fake_proxy(
        ...    compute_service.ComputeService,
        ... )
        >>> # configure return values for various proxy APIs
        >>> # note that this will generate new fake resources on each invocation
        >>> fake_compute_proxy.get_server.side_effect = functools.partial(
        ...     fakes.generate_fake_resource,
        ...     server.Server,
        ... )
        >>> fake_compute_proxy.servers.side_effect = functools.partial(
        ...     fakes.generate_fake_resources,
        ...     server.Server,
        ... )
        >>> fake_compute_proxy.servers()
        <generator object generate_fake_resources at 0x7f92768dc040>
        >>> fake_compute_proxy.serverssss()
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File "/usr/lib64/python3.11/unittest/mock.py", line 653, in __getattr__
            raise AttributeError("Mock object has no attribute %r" % name)
        AttributeError: Mock object has no attribute 'serverssss'. Did you mean: 'server_ips'?

    :param service: The service to generate the fake proxy for.
    :type service: :class:`~openstack.service_description.ServiceDescription`
    :param api_version: The API version to generate the fake proxy for.
        This should be a major version must be supported by openstacksdk, as
        specified in the ``supported_versions`` attribute of the provided
        ``service``. This is only required if openstacksdk supports multiple
        API versions for the given service.
    :type api_version: int or None
    :raises ValueError: if the ``service`` is not a valid
        :class:`~openstack.service_description.ServiceDescription` or if
        ``api_version`` is not supported
    :returns: An autospecced mock of the :class:`~openstack.proxy.Proxy`
        implementation for the specified service type and API version
    """
    if not issubclass(service, service_description.ServiceDescription):
        raise ValueError(
            f"Service {service.__name__} is not a valid ServiceDescription"
        )

    supported_versions = service.supported_versions

    if api_version is None:
        if len(supported_versions) > 1:
            raise ValueError(
                f"api_version was not provided but service {service.__name__} "
                f"provides multiple API versions"
            )
        else:
            api_version = list(supported_versions)[0]
    elif api_version not in supported_versions:
        raise ValueError(
            f"API version {api_version} is not supported by openstacksdk. "
            f"Supported API versions are: {', '.join(supported_versions)}"
        )

    return mock.create_autospec(supported_versions[api_version])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import vpn_ike_policy
from openstack.tests.functional import base


class TestVpnIkePolicy(base.BaseFunctionalTest):
    ID = None

    def setUp(self):
        super(TestVpnIkePolicy, self).setUp()
        if not self.user_cloud._has_neutron_extension("vpnaas"):
            self.skipTest("vpnaas service not supported by cloud")
        self.IKEPOLICY_NAME = self.getUniqueString("ikepolicy")
        self.UPDATE_NAME = self.getUniqueString("ikepolicy-updated")
        policy = self.user_cloud.network.create_vpn_ike_policy(
            name=self.IKEPOLICY_NAME
        )
        assert isinstance(policy, vpn_ike_policy.VpnIkePolicy)
        self.assertEqual(self.IKEPOLICY_NAME, policy.name)
        self.ID = policy.id

    def tearDown(self):
        ikepolicy = self.user_cloud.network.delete_vpn_ike_policy(
            self.ID, ignore_missing=True
        )
        self.assertIsNone(ikepolicy)
        super(TestVpnIkePolicy, self).tearDown()

    def test_list(self):
        policies = [f.name for f in self.user_cloud.network.vpn_ike_policies()]
        self.assertIn(self.IKEPOLICY_NAME, policies)

    def test_find(self):
        policy = self.user_cloud.network.find_vpn_ike_policy(
            self.IKEPOLICY_NAME
        )
        self.assertEqual(self.ID, policy.id)

    def test_get(self):
        policy = self.user_cloud.network.get_vpn_ike_policy(self.ID)
        self.assertEqual(self.IKEPOLICY_NAME, policy.name)
        self.assertEqual(self.ID, policy.id)

    def test_update(self):
        policy = self.user_cloud.network.update_vpn_ike_policy(
            self.ID, name=self.UPDATE_NAME
        )
        self.assertEqual(self.UPDATE_NAME, policy.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class CronTrigger(resource.Resource):
    resource_key = 'cron_trigger'
    resources_key = 'cron_triggers'
    base_path = '/cron_triggers'

    # capabilities
    allow_create = True
    allow_list = True
    allow_fetch = True
    allow_delete = True

    _query_mapping = resource.QueryParameters(
        'marker',
        'limit',
        'sort_keys',
        'sort_dirs',
        'fields',
        'name',
        'workflow_name',
        'workflow_id',
        'workflow_input',
        'workflow_params',
        'scope',
        'pattern',
        'remaining_executions',
        'project_id',
        'first_execution_time',
        'next_execution_time',
        'created_at',
        'updated_at',
        'all_projects',
    )

    #: The name of this Cron Trigger
    name = resource.Body("name")
    #: The pattern for this Cron Trigger
    pattern = resource.Body("pattern")
    #: Count of remaining exectuions
    remaining_executions = resource.Body("remaining_executions")
    #: Time of the first execution
    first_execution_time = resource.Body("first_execution_time")
    #: Time of the next execution
    next_execution_time = resource.Body("next_execution_time")
    #: Workflow name
    workflow_name = resource.Body("workflow_name")
    #: Workflow ID
    workflow_id = resource.Body("workflow_id")
    #: The inputs for Workflow
    workflow_input = resource.Body("workflow_input")
    #: Workflow params
    workflow_params = resource.Body("workflow_params")
    #: The ID of the associated project
    project_id = resource.Body("project_id")
    #: The time at which the cron trigger was created
    created_at = resource.Body("created_at")
    #: The time at which the cron trigger was created
    updated_at = resource.Body("updated_at")

    def create(self, session, base_path=None):
        return super(CronTrigger, self).create(
            session, prepend_key=False, base_path=base_path
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack.common import tag
from openstack import resource
from openstack import utils


class QoSPolicy(resource.Resource, tag.TagMixin):
    resource_key = 'policy'
    resources_key = 'policies'
    base_path = '/qos/policies'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'name',
        'description',
        'is_default',
        'project_id',
        is_shared='shared',
        **tag.TagMixin._tag_query_parameters
    )

    # Properties
    #: QoS policy name.
    name = resource.Body('name')
    #: The ID of the project who owns the network. Only administrative
    #: users can specify a project ID other than their own.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The QoS policy description.
    description = resource.Body('description')
    #: Indicates whether this QoS policy is the default policy for this
    #: project.
    #: *Type: bool*
    is_default = resource.Body('is_default', type=bool)
    #: Indicates whether this QoS policy is shared across all projects.
    #: *Type: bool*
    is_shared = resource.Body('shared', type=bool)
    #: List of QoS rules applied to this QoS policy.
    rules = resource.Body('rules')

    def set_tags(self, session, tags):
        url = utils.urljoin('/policies', self.id, 'tags')
        session.put(url, json={'tags': tags})
        self._body.attributes.update({'tags': tags})
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_volume
----------------------------------

Functional tests for block storage methods.
"""

from fixtures import TimeoutException
from testtools import content

from openstack import exceptions
from openstack.tests.functional import base
from openstack import utils


class TestVolume(base.BaseFunctionalTest):
    # Creating and deleting volumes is slow
    TIMEOUT_SCALING_FACTOR = 1.5

    def setUp(self):
        super(TestVolume, self).setUp()
        self.skipTest('Volume functional tests temporarily disabled')
        if not self.user_cloud.has_service('volume'):
            self.skipTest('volume service not supported by cloud')

    def test_volumes(self):
        '''Test volume and snapshot functionality'''
        volume_name = self.getUniqueString()
        snapshot_name = self.getUniqueString()
        self.addDetail('volume', content.text_content(volume_name))
        self.addCleanup(self.cleanup, volume_name, snapshot_name=snapshot_name)
        volume = self.user_cloud.create_volume(
            display_name=volume_name, size=1
        )
        snapshot = self.user_cloud.create_volume_snapshot(
            volume['id'], display_name=snapshot_name
        )

        ret_volume = self.user_cloud.get_volume_by_id(volume['id'])
        self.assertEqual(volume['id'], ret_volume['id'])

        volume_ids = [v['id'] for v in self.user_cloud.list_volumes()]
        self.assertIn(volume['id'], volume_ids)

        snapshot_list = self.user_cloud.list_volume_snapshots()
        snapshot_ids = [s['id'] for s in snapshot_list]
        self.assertIn(snapshot['id'], snapshot_ids)

        ret_snapshot = self.user_cloud.get_volume_snapshot_by_id(
            snapshot['id']
        )
        self.assertEqual(snapshot['id'], ret_snapshot['id'])

        self.user_cloud.delete_volume_snapshot(snapshot_name, wait=True)
        self.user_cloud.delete_volume(volume_name, wait=True)

    def test_volume_to_image(self):
        '''Test volume export to image functionality'''
        volume_name = self.getUniqueString()
        image_name = self.getUniqueString()
        self.addDetail('volume', content.text_content(volume_name))
        self.addCleanup(self.cleanup, volume_name, image_name=image_name)
        volume = self.user_cloud.create_volume(
            display_name=volume_name, size=1
        )
        image = self.user_cloud.create_image(
            image_name, volume=volume, wait=True
        )

        volume_ids = [v['id'] for v in self.user_cloud.list_volumes()]
        self.assertIn(volume['id'], volume_ids)

        image_list = self.user_cloud.list_images()
        image_ids = [s['id'] for s in image_list]
        self.assertIn(image['id'], image_ids)

        self.user_cloud.delete_image(image_name, wait=True)
        self.user_cloud.delete_volume(volume_name, wait=True)

    def cleanup(self, volume, snapshot_name=None, image_name=None):
        # Need to delete snapshots before volumes
        if snapshot_name:
            snapshot = self.user_cloud.get_volume_snapshot(snapshot_name)
            if snapshot:
                self.user_cloud.delete_volume_snapshot(
                    snapshot_name, wait=True
                )
        if image_name:
            image = self.user_cloud.get_image(image_name)
            if image:
                self.user_cloud.delete_image(image_name, wait=True)
        if not isinstance(volume, list):
            self.user_cloud.delete_volume(volume, wait=True)
        else:
            # We have more than one volume to clean up - submit all of the
            # deletes without wait, then poll until none of them are found
            # in the volume list anymore
            for v in volume:
                self.user_cloud.delete_volume(v, wait=False)
            try:
                for count in utils.iterate_timeout(
                    180, "Timeout waiting for volume cleanup"
                ):
                    found = False
                    for existing in self.user_cloud.list_volumes():
                        for v in volume:
                            if v['id'] == existing['id']:
                                found = True
                                break
                        if found:
                            break
                    if not found:
                        break
            except (exceptions.ResourceTimeout, TimeoutException):
                # NOTE(slaweq): ups, some volumes are still not removed
                # so we should try to force delete it once again and move
                # forward
                for existing in self.user_cloud.list_volumes():
                    for v in volume:
                        if v['id'] == existing['id']:
                            self.operator_cloud.delete_volume(
                                v, wait=False, force=True
                            )

    def test_list_volumes_pagination(self):
        '''Test pagination for list volumes functionality'''

        volumes = []
        # the number of created volumes needs to be higher than
        # CONF.osapi_max_limit but not higher than volume quotas for
        # the test user in the tenant(default quotas is set to 10)
        num_volumes = 8
        for i in range(num_volumes):
            name = self.getUniqueString()
            v = self.user_cloud.create_volume(display_name=name, size=1)
            volumes.append(v)
        self.addCleanup(self.cleanup, volumes)
        result = []
        for i in self.user_cloud.list_volumes():
            if i['name'] and i['name'].startswith(self.id()):
                result.append(i['id'])
        self.assertEqual(sorted([i['id'] for i in volumes]), sorted(result))

    def test_update_volume(self):
        name, desc = self.getUniqueString('name'), self.getUniqueString('desc')
        self.addCleanup(self.cleanup, name)
        volume = self.user_cloud.create_volume(1, name=name, description=desc)
        self.assertEqual(volume.name, name)
        self.assertEqual(volume.description, desc)
        new_name = self.getUniqueString('name')
        volume = self.user_cloud.update_volume(volume.id, name=new_name)
        self.assertNotEqual(volume.name, name)
        self.assertEqual(volume.name, new_name)
        self.assertEqual(volume.description, desc)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack.common import tag
from openstack import exceptions
from openstack.network.v2 import _base
from openstack import resource
from openstack import utils


class Router(_base.NetworkResource, tag.TagMixin):
    resource_key = 'router'
    resources_key = 'routers'
    base_path = '/routers'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # NOTE: We don't support query on datetime, list or dict fields
    _query_mapping = resource.QueryParameters(
        'description',
        'flavor_id',
        'name',
        'status',
        'project_id',
        is_admin_state_up='admin_state_up',
        is_distributed='distributed',
        is_ha='ha',
        **tag.TagMixin._tag_query_parameters
    )

    # Properties
    #: Availability zone hints to use when scheduling the router.
    #: *Type: list of availability zone names*
    availability_zone_hints = resource.Body(
        'availability_zone_hints', type=list
    )
    #: Availability zones for the router.
    #: *Type: list of availability zone names*
    availability_zones = resource.Body('availability_zones', type=list)
    #: Timestamp when the router was created.
    created_at = resource.Body('created_at')
    #: The router description.
    description = resource.Body('description')
    #: The ndp proxy state of the router
    enable_ndp_proxy = resource.Body('enable_ndp_proxy', type=bool)
    #: The ``network_id``, for the external gateway. *Type: dict*
    external_gateway_info = resource.Body('external_gateway_info', type=dict)
    #: The ID of the flavor.
    flavor_id = resource.Body('flavor_id')
    #: The administrative state of the router, which is up ``True``
    #: or down ``False``. *Type: bool*
    is_admin_state_up = resource.Body('admin_state_up', type=bool)
    #: The distributed state of the router, which is distributed ``True``
    #: or not ``False``. *Type: bool*
    is_distributed = resource.Body('distributed', type=bool)
    #: The highly-available state of the router, which is highly available
    #: ``True`` or not ``False``. *Type: bool*
    is_ha = resource.Body('ha', type=bool)
    #: The router name.
    name = resource.Body('name')
    #: The ID of the project this router is associated with.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: Revision number of the router. *Type: int*
    revision_number = resource.Body('revision', type=int)
    #: The extra routes configuration for the router.
    routes = resource.Body('routes', type=list)
    #: The router status.
    status = resource.Body('status')
    #: Timestamp when the router was created.
    updated_at = resource.Body('updated_at')

    def _put(self, session, url, body):
        resp = session.put(url, json=body)
        exceptions.raise_from_response(resp)
        return resp

    def add_interface(self, session, **body):
        """Add an internal interface to a logical router.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param dict body: The body requested to be updated on the router

        :returns: The body of the response as a dictionary.

        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'add_router_interface')
        resp = self._put(session, url, body)
        return resp.json()

    def remove_interface(self, session, **body):
        """Remove an internal interface from a logical router.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param dict body: The body requested to be updated on the router

        :returns: The body of the response as a dictionary.

        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'remove_router_interface')
        resp = self._put(session, url, body)
        return resp.json()

    def add_extra_routes(self, session, body) -> 'Router':
        """Add extra routes to a logical router.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param dict body: The request body as documented in the api-ref.

        :returns: The response as a Router object with the added extra routes.

        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'add_extraroutes')
        resp = self._put(session, url, body)
        self._translate_response(resp)
        return self

    def remove_extra_routes(self, session, body) -> 'Router':
        """Remove extra routes from a logical router.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param dict body: The request body as documented in the api-ref.

        :returns: The response as a Router object with the extra routes left.

        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'remove_extraroutes')
        resp = self._put(session, url, body)
        self._translate_response(resp)
        return self

    def add_gateway(self, session, **body):
        """Add an external gateway to a logical router.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param dict body: The body requested to be updated on the router

        :returns: The body of the response as a dictionary.
        """
        url = utils.urljoin(self.base_path, self.id, 'add_gateway_router')
        resp = session.put(url, json=body)
        return resp.json()

    def remove_gateway(self, session, **body):
        """Remove an external gateway from a logical router.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param dict body: The body requested to be updated on the router

        :returns: The body of the response as a dictionary.
        """
        url = utils.urljoin(self.base_path, self.id, 'remove_gateway_router')
        resp = session.put(url, json=body)
        return resp.json()

    def add_external_gateways(self, session, body):
        """Add external gateways to a router.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param dict body: The body requested to be updated on the router

        :returns: The body of the response as a dictionary.
        """
        url = utils.urljoin(self.base_path, self.id, 'add_external_gateways')
        resp = session.put(url, json=body)
        self._translate_response(resp)
        return self

    def update_external_gateways(self, session, body):
        """Update external gateways of a router.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param dict body: The body requested to be updated on the router

        :returns: The body of the response as a dictionary.
        """
        url = utils.urljoin(
            self.base_path, self.id, 'update_external_gateways'
        )
        resp = session.put(url, json=body)
        self._translate_response(resp)
        return self

    def remove_external_gateways(self, session, body):
        """Remove external gateways from a router.

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param dict body: The body requested to be updated on the router

        :returns: The body of the response as a dictionary.
        """
        url = utils.urljoin(
            self.base_path, self.id, 'remove_external_gateways'
        )
        resp = session.put(url, json=body)
        self._translate_response(resp)
        return self


class L3AgentRouter(Router):
    resource_key = 'router'
    resources_key = 'routers'
    base_path = '/agents/%(agent_id)s/l3-routers'
    resource_name = 'l3-router'

    # capabilities
    allow_create = False
    allow_retrieve = True
    allow_commit = False
    allow_delete = False
    allow_list = True


# NOTE: No query parameter is supported



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.compute.v2 import keypair
from openstack.tests.functional import base


class TestKeypair(base.BaseFunctionalTest):
    def setUp(self):
        super(TestKeypair, self).setUp()

        # Keypairs can't have .'s in the name. Because why?
        self.NAME = self.getUniqueString().split('.')[-1]

        sot = self.conn.compute.create_keypair(name=self.NAME, type='ssh')
        assert isinstance(sot, keypair.Keypair)
        self.assertEqual(self.NAME, sot.name)
        self._keypair = sot

    def tearDown(self):
        sot = self.conn.compute.delete_keypair(self._keypair)
        self.assertIsNone(sot)
        super(TestKeypair, self).tearDown()

    def test_find(self):
        sot = self.conn.compute.find_keypair(self.NAME)
        self.assertEqual(self.NAME, sot.name)
        self.assertEqual(self.NAME, sot.id)

    def test_get(self):
        sot = self.conn.compute.get_keypair(self.NAME)
        self.assertEqual(self.NAME, sot.name)
        self.assertEqual(self.NAME, sot.id)
        self.assertEqual('ssh', sot.type)

    def test_list(self):
        names = [o.name for o in self.conn.compute.keypairs()]
        self.assertIn(self.NAME, names)


class TestKeypairAdmin(base.BaseFunctionalTest):
    def setUp(self):
        super(TestKeypairAdmin, self).setUp()
        self._set_operator_cloud(interface='admin')

        self.NAME = self.getUniqueString().split('.')[-1]
        self.USER = self.operator_cloud.list_users()[0]

        sot = self.conn.compute.create_keypair(
            name=self.NAME, user_id=self.USER.id
        )
        assert isinstance(sot, keypair.Keypair)
        self.assertEqual(self.NAME, sot.name)
        self.assertEqual(self.USER.id, sot.user_id)
        self._keypair = sot

    def tearDown(self):
        sot = self.conn.compute.delete_keypair(self._keypair)
        self.assertIsNone(sot)
        super(TestKeypairAdmin, self).tearDown()

    def test_get(self):
        sot = self.conn.compute.get_keypair(self.NAME)
        self.assertEqual(self.NAME, sot.name)
        self.assertEqual(self.NAME, sot.id)
        self.assertEqual(self.USER.id, sot.user_id)
